"""Dropout layers in `tf.keras`."""

import math

import tensorflow as tf
import tensorflow_probability as tfp

tfd = tfp.distributions
tfk = tf.keras
tfkl = tfk.layers

# TODO: add `K.in_train_phase`.


class BernoulliDropout(tfkl.Layer):
    """Bernoulli dropout layer.

    Outputs the input element multiplied by a random variable
    sampled from a Bernoulli distribution with either mean keep_prob
    (scale_during_training False) or mean 1 (scale_during_training True)

    Parameters
    ----------
    rate : float between 0 and 1, drop probability.
    is_monte_carlo : A boolean Tensor corresponding to whether or not Monte-Carlo
        sampling will be used to calculate the networks output
    scale_during_training : A boolean value determining whether scaling is performed
        during training or testing
    seed : int, value to seed random number generator.

    References
    ----------
    Dropout: A Simple Way to Prevent Neural Networks from Overfitting.
        N. Srivastava, G. Hinton, A. Krizhevsky, I. Sutskever & R. Salakhutdinov,
        (2014), Journal of Machine Learning Research, 5(Jun)(2), 1929-1958.

    Links
    -----
    [https://www.cs.toronto.edu/~hinton/absps/JMLRdropout.pdf](https://www.cs.toronto.edu/~hinton/absps/JMLRdropout.pdf)
    """

    def __init__(
        self, rate, is_monte_carlo, scale_during_training=True, seed=None, **kwargs
    ):
        super().__init__(**kwargs)
        self.rate = rate
        self.is_monte_carlo = is_monte_carlo
        self.scale_during_training = scale_during_training
        self.keep_prob = 1.0 - rate
        self.seed = seed

    def call(self, x):
        if self.is_monte_carlo:
            d = tf.nn.dropout(x, rate=self.rate, seed=self.seed)
            return d if self.scale_during_training else d * self.keep_prob
        return x if self.scale_during_training else self.keep_prob * x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "rate": self.rate,
                "is_monte_carlo": self.is_monte_carlo,
                "scale_during_training": self.scale_during_training,
                "seed": self.seed,
            }
        )
        return config


def divergence_fn(pl, pr, scale_factor):
    """Divergence computation for concrete dropout"""
    return tf.reduce_sum(
        tf.add(
            tf.multiply(
                pl,
                tf.subtract(
                    tf.math.log(tf.add(pl, tfk.backend.epsilon())),
                    tf.math.log(pr),
                ),
            ),
            tf.multiply(
                tf.subtract(tfk.backend.constant(1), pl),
                tf.subtract(
                    tf.math.log(
                        tf.add(
                            tf.subtract(tfk.backend.constant(1), pl),
                            tfk.backend.epsilon(),
                        )
                    ),
                    tf.math.log(pr),
                ),
            ),
        )
    ) / tf.cast(scale_factor, dtype=tf.float32)


class ConcreteDropout(tfkl.Layer):
    """Concrete Dropout.
    Outputs the input element multiplied by a random variable sampled from a concrete
    distribution
    Parameters
    ----------
    is_monte_carlo : A boolean Tensor corresponding to whether or not Monte-Carlo
        sampling will be used to calculate the networks output temperature.
    use_expectation : boolean
    seed : int, value to seed random number generator.
    name : A name for this layer (optional).
    References
    ----------
    Concrete Dropout. Y. Gal, J. Hron & A. Kendall, Advances in Neural Information
    Processing Systems. 2017.
    Links
    -----
    [http://papers.nips.cc/paper/6949-concrete-dropout.pdf](http://papers.nips.cc/paper/6949-concrete-dropout.pdf)
    """

    def __init__(
        self,
        is_monte_carlo=False,
        temperature=0.02,
        use_expectation=False,
        scale_factor=1,
        seed=None,
        **kwargs
    ):
        super(ConcreteDropout, self).__init__(**kwargs)
        self.is_monte_carlo = is_monte_carlo
        self.temperature = temperature
        self.use_expectation = use_expectation
        self.seed = seed
        self.scale_factor = scale_factor

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape)
        # if self.concrete:
        p_prior = tfk.initializers.Constant(0.5)
        initial_p = tfk.initializers.Constant(0.9)
        self.p_post = self.add_variable(
            name="p_l", shape=input_shape[-1:], initializer=initial_p, trainable=True
        )

        self.p_prior = self.add_variable(
            name="pr", shape=input_shape[-1:], initializer=p_prior, trainable=False
        )
        self.p_post = tfk.backend.clip(self.p_post, 0.05, 0.95)
        self.built = True

    def call(self, x):
        outputs = self._apply_concrete(x)
        self._apply_divergence_concrete(self.scale_factor, name="concrete_loss")
        return outputs

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "is_monte_carlo": self.is_monte_carlo,
                "temperature": self.temperature,
                "use_expectation": self.use_expectation,
                "seed": self.seed,
                "scale": self.scale_factor,
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        config = config.copy()
        return cls(**config)

    def _apply_concrete(self, inp):
        inference = inp
        eps = tfk.backend.epsilon()
        use_expectation = self.use_expectation
        if self.is_monte_carlo:
            noise = tf.random.uniform(
                tf.shape(inference),
                minval=0,
                maxval=1,
                seed=self.seed,
                dtype=self.dtype,
            )
            z = tf.nn.sigmoid(
                (
                    tf.math.log(self.p_post + eps)
                    - tf.math.log(1.0 - self.p_post + eps)
                    + tf.math.log(noise + eps)
                    - tf.math.log(1.0 - noise + eps)
                )
                / self.temperature
            )
            return inp * z
        else:
            return inference * self.p_post if use_expectation else inference

    def _apply_divergence_concrete(self, scale_factor, name):
        divergence = tf.identity(
            divergence_fn(self.p_post, self.p_prior, scale_factor), name=name
        )
        self.add_loss(divergence)


class GaussianDropout(tfkl.Layer):
    """Gaussian Dropout.

    Parameters
    ----------
    rate : float between 0 and 1, drop probability.
    is_monte_carlo : A boolean Tensor corresponding to whether or not Monte-Carlo
        sampling will be used to calculate the networks output
    scale_during_training : A boolean value determining whether scaling is performed
        during training or testing
    seed : int, value to seed random number generator.

    References
    ----------
    Dropout: A Simple Way to Prevent Neural Networks from Overfitting.
    N. Srivastava, G. Hinton, A. Krizhevsky, I. Sutskever & R. Salakhutdinov,
    (2014), Journal of Machine Learning Research, 5(Jun)(2), 1929-1958.

    Links
    -----
    [https://www.cs.toronto.edu/~hinton/absps/JMLRdropout.pdf](https://www.cs.toronto.edu/~hinton/absps/JMLRdropout.pdf)
    """

    def __init__(
        self, rate, is_monte_carlo, scale_during_training=True, seed=None, **kwargs
    ):
        super().__init__(**kwargs)
        self.rate = rate
        self.is_monte_carlo = is_monte_carlo
        self.scale_during_training = scale_during_training
        self.seed = seed

    def call(self, x):
        if self.is_monte_carlo:
            if self.scale_during_training:
                stddev = math.sqrt(self.rate / (1.0 - self.rate))
            else:
                stddev = math.sqrt(self.rate * (1.0 - self.rate))
            noise = tf.random.normal(
                tf.shape(x), mean=1.0, stddev=stddev, seed=self.seed
            )
            return x * noise
        else:
            return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "rate": self.rate,
                "is_monte_carlo": self.is_monte_carlo,
                "scale_during_training": self.scale_during_training,
                "seed": self.seed,
            }
        )
        return config



"""Main command-line interface for nobrainer."""

import datetime
import glob
import json
import logging
import os
import platform
import sys

import click
import nibabel as nib
import numpy as np
import skimage.measure
import skimage.transform
import tensorflow as tf
import tensorflow_probability as tfp

from .. import __version__
from ..io import read_csv as _read_csv
from ..io import read_volume as _read_volume
from ..io import verify_features_labels as _verify_features_labels
from ..prediction import _transform_and_predict
from ..tfrecord import write as _write_tfrecord
from ..utils import get_num_parallel
from ..volume import from_blocks_numpy as _from_blocks_numpy
from ..volume import standardize_numpy as _standardize_numpy
from ..volume import to_blocks_numpy as _to_blocks_numpy

_option_kwds = {"show_default": True}


class JSONParamType(click.ParamType):
    name = "json"

    def convert(self, value, param, ctx):
        try:
            return json.loads(value)
        except json.decoder.JSONDecodeError:
            self.fail("%s is not valid JSON" % value, param, ctx)


@click.group()
@click.version_option(__version__, message="%(prog)s version %(version)s")
def cli():
    """A framework for developing neural network models for 3D image processing."""
    return


@cli.command()
@click.option(
    "-c", "--csv", type=click.Path(exists=True), required=True, **_option_kwds
)
@click.option(
    "-t",
    "--tfrecords-template",
    default="tfrecords/data_shard-{shard:03d}.tfrec",
    required=True,
    **_option_kwds,
)
@click.option("-s", "--volume-shape", nargs=3, type=int, required=True, **_option_kwds)
@click.option(
    "-n",
    "--examples-per-shard",
    type=int,
    default=100,
    help="Number of (feature, label) pairs per TFRecord file.",
    **_option_kwds,
)
@click.option(
    "--to-ras/--no-to-ras",
    default=True,
    help="Reorient volumes to RAS before saving to TFRecords.",
    **_option_kwds,
)
@click.option(
    "--gzip/--no-gzip",
    default=True,
    help="Compress TFRecords with gzip (highly recommended).",
    **_option_kwds,
)
@click.option(
    "--verify-volumes/--no-verify-volumes",
    default=True,
    help=(
        "Verify volume pairs before converting. This option is highly recommended, as"
        ' it checks that shapes of features and labels are equal to "volume-shape",'
        " that labels are (or can safely be coerced to) an integer type, and that"
        " labels are all >= 0."
    ),
    **_option_kwds,
)
@click.option(
    "-j",
    "--num-parallel-calls",
    default=-1,
    type=int,
    help="Number of processes to use. If -1, uses all available processes.",
    **_option_kwds,
)
@click.option(
    "--multi-resolution/--no-multi-resolution",
    default=False,
    help="Create tfrecords for multiple resolutions. if set, labels in csv need to be scalar.",
    **_option_kwds,
)
@click.option(
    "--start-resolution",
    type=int,
    default=4,
    help=(
        "Set if `multi-resolution` is true. Indicates smallest resolution for "
        "tfrecords, all resolutions in exponents of 2 from `start-resolution` to"
        " `volume-shape` are generated as tfrecords"
    ),
    **_option_kwds,
)
@click.option(
    "-v", "--verbose", is_flag=True, help="Print progress bar.", **_option_kwds
)
def convert(
    *,
    csv,
    tfrecords_template,
    volume_shape,
    examples_per_shard,
    to_ras,
    gzip,
    verify_volumes,
    num_parallel_calls,
    multi_resolution,
    start_resolution,
    verbose,
):
    """Convert medical imaging volumes to TFRecords.

    Volumes must all be the same shape. This will overwrite existing TFRecord files.

    Labels can be volumetric or scalar.
    """
    # TODO: improve docs.
    volume_filepaths = _read_csv(csv)
    num_parallel_calls = (
        get_num_parallel() if num_parallel_calls == -1 else num_parallel_calls
    )

    _dirname = os.path.dirname(tfrecords_template)
    if not os.path.exists(_dirname):
        raise ValueError("directory does not exist: {}".format(_dirname))

    if verify_volumes:
        invalid_pairs = _verify_features_labels(
            volume_filepaths=volume_filepaths,
            volume_shape=volume_shape,
            check_shape=True,
            check_labels_int=True,
            check_labels_gte_zero=True,
            num_parallel_calls=None,
            verbose=1,
        )

        if not invalid_pairs:
            click.echo(click.style("Passed verification.", fg="green"))
        else:
            click.echo(click.style("Failed verification.", fg="red"))
            click.echo(
                f"Found {len(invalid_pairs)} invalid pairs of volumes. These files"
                " might not all have shape {volume_shape}, the labels might not be an"
                " integer type or coercible to integer type, or the labels might not"
                " be >= 0."
            )
            for pair in invalid_pairs:
                click.echo(pair[0])
                click.echo(pair[1])
            sys.exit(-1)

    if multi_resolution:
        start_resolution_log = np.log2(start_resolution).astype(np.int32)
        target_resolution_log = np.log2(volume_shape[0]).astype(np.int32)
        resolutions = [
            2**res for res in range(start_resolution_log, target_resolution_log + 1)
        ]
    else:
        resolutions = None

    _write_tfrecord(
        features_labels=volume_filepaths,
        filename_template=tfrecords_template,
        examples_per_shard=examples_per_shard,
        to_ras=to_ras,
        compressed=gzip,
        processes=num_parallel_calls,
        multi_resolution=multi_resolution,
        resolutions=resolutions,
        verbose=verbose,
    )

    click.echo(click.style("Finished conversion to TFRecords.", fg="green"))


@cli.command()
def merge():
    """Merge multiple models trained with variational weights.

    These models must have the same architecture and should have been trained
    from the same initial model.
    """
    click.echo(
        "Not implemented yet. In the future, this command will be used for merging"
        " models."
    )
    sys.exit(-2)


@cli.command()
@click.argument("infile")
@click.argument("outfile")
@click.option(
    "-m",
    "--model",
    type=click.Path(exists=True),
    required=True,
    help="Path to model HDF5 file.",
    **_option_kwds,
)
@click.option(
    "-b",
    "--block-shape",
    default=(128, 128, 128),
    type=int,
    nargs=3,
    help="Shape of sub-volumes on which to predict.",
    **_option_kwds,
)
@click.option(
    "-r",
    "--resize-features-to",
    default=(256, 256, 256),
    type=int,
    nargs=3,
    help="Resize features to this size before taking blocks and predicting.",
    **_option_kwds,
)
@click.option(
    "-t",
    "--threshold",
    type=float,
    default=0.3,
    help=(
        "Threshold used to binarize model output. Only used in binary prediction and"
        " must be in (0, 1)."
    ),
    **_option_kwds,
)
@click.option(
    "-l",
    "--largest-label",
    is_flag=True,
    help=(
        "Zero out all values not connected to the largest contiguous label (not"
        " including 0 values). This remove false positives in binary prediction."
    ),
    **_option_kwds,
)
@click.option(
    "--rotate-and-predict",
    is_flag=True,
    help=(
        "Average the prediction with a prediction on a rotated (and subsequently"
        " un-rotated) volume. This can produce a better overall prediction."
    ),
    **_option_kwds,
)
@click.option(
    "-v", "--verbose", is_flag=True, help="Print progress bar.", **_option_kwds
)
def predict(
    *,
    infile,
    outfile,
    model,
    block_shape,
    resize_features_to,
    threshold,
    largest_label,
    rotate_and_predict,
    verbose,
):
    """Predict labels from features using a trained model.

    The predictions are saved to OUTFILE.
    """

    if not verbose:
        # Suppress most logging messages.
        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
        tf.get_logger().setLevel(logging.ERROR)

    if os.path.exists(outfile):
        raise FileExistsError(
            "Output file already exists. Will not overwrite {}".format(outfile)
        )

    x, affine = _read_volume(infile, dtype=np.float32, return_affine=True)
    if x.ndim != 3:
        raise ValueError("Input volume must be rank 3, got rank {}".format(x.ndim))
    original_shape = x.shape
    required_shape = resize_features_to
    must_resize = False
    if x.shape != required_shape:
        must_resize = True
        if verbose:
            click.echo(
                "Resizing volume from shape {} to shape {}".format(
                    x.shape, required_shape
                )
            )
        x = skimage.transform.resize(
            x,
            output_shape=required_shape,
            order=1,  # linear
            mode="constant",
            preserve_range=True,
            anti_aliasing=False,
        )

    x = _standardize_numpy(x)
    x_blocks = _to_blocks_numpy(x, block_shape=block_shape)
    x_blocks = x_blocks[..., None]  # Add grayscale channel.

    model = tf.keras.models.load_model(model, compile=False)
    if verbose:
        click.echo("Predicting ...")
    try:
        y_blocks = model.predict(x_blocks, batch_size=1, verbose=verbose)
    except Exception:
        click.echo(click.style("ERROR: prediction failed. See error trace.", fg="red"))
        raise

    # Collapse the last dimension, depending on number of output classes.
    is_binary_prediction = y_blocks.shape[-1] == 1
    if is_binary_prediction:
        y_blocks = y_blocks.squeeze(-1)
    else:
        y_blocks = y_blocks.argmax(-1)

    y = _from_blocks_numpy(y_blocks, x.shape)

    # Rotate the volume, predict, undo the rotation, and average with original
    # prediction.
    if rotate_and_predict:
        if not is_binary_prediction:
            raise ValueError("Cannot transform and predict on multi-class output.")
        if verbose:
            click.echo("Predicting on rotated volume ...")
        y_other = _transform_and_predict(
            model=model,
            x=x,
            block_shape=block_shape,
            rotation=[np.pi / 4, np.pi / 4, 0],
            translation=[0, 0, 0],
            verbose=verbose,
        )
        if verbose:
            click.echo("Averaging predictions ...")
        y = np.mean([y, y_other], axis=0)

    if is_binary_prediction:
        if threshold <= 0 or threshold >= 1:
            raise ValueError("Threshold must be in (0, 1).")
        y = y > threshold

    if must_resize:
        if verbose:
            click.echo(
                "Resizing volume from shape {} to shape {}".format(
                    y.shape, original_shape
                )
            )
        y = skimage.transform.resize(
            y,
            output_shape=original_shape,
            order=0,  # nearest neighbor
            mode="constant",
            preserve_range=True,
            anti_aliasing=False,
        )

    if largest_label:
        if not is_binary_prediction:
            raise ValueError(
                "Removing all labels except the largest is only allowed with binary"
                " prediction."
            )
        if verbose:
            click.echo("Removing all labels except largest ...")
        labels, n_labels = skimage.measure.label(y, return_num=True)
        # Do not consider 0 values.
        d = {(labels == label).sum(): label for label in range(1, n_labels + 1)}
        largest_label = d[max(d.keys())]
        if verbose:
            click.echo(
                "Zeroed {} region(s) not contiguous with largest label.".format(
                    n_labels - 2
                )
            )
        y = (labels == largest_label).astype(np.int32)

    img = nib.spatialimages.SpatialImage(y.astype(np.int32), affine=affine)
    nib.save(img, outfile)
    if verbose:
        click.echo("Output saved to {}".format(outfile))


@cli.command()
@click.argument("outfile")
@click.option(
    "-m",
    "--model",
    type=click.Path(exists=True),
    required=True,
    help="Path to saved models directory containing the HDF5 files.",
    **_option_kwds,
)
@click.option(
    "-l",
    "--latent-size",
    type=int,
    default=1024,
    help=("Input latent size for the generator."),
    **_option_kwds,
)
@click.option(
    "--drange-in",
    default=(-1, 1),
    type=int,
    nargs=2,
    help="Range of values of image generated by model.",
    **_option_kwds,
)
@click.option(
    "--drange-out",
    default=(0, 255),
    type=int,
    nargs=2,
    help="Desired output range of values of image.",
    **_option_kwds,
)
@click.option(
    "-o",
    "--output-shape",
    default=(128, 128, 128),
    type=int,
    nargs=3,
    help="Shape of sub-volumes to generate.",
    **_option_kwds,
)
@click.option(
    "--multi-resolution",
    is_flag=True,
    help=(
        "Generate all resolutions available from the progressive generator using the same latents."
    ),
    **_option_kwds,
)
@click.option(
    "-v", "--verbose", is_flag=True, help="Print progress bar.", **_option_kwds
)
def generate(
    *,
    outfile,
    model,
    latent_size,
    drange_in,
    drange_out,
    output_shape,
    multi_resolution,
    verbose,
):
    """Generate images from latents using a trained GAN model.

    The generated image is saved to OUTFILE.
    """

    if not verbose:
        # Suppress most logging messages.
        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
        tf.get_logger().setLevel(logging.ERROR)

    if os.path.exists(outfile):
        raise FileExistsError(
            "Output file already exists. Will not overwrite {}".format(outfile)
        )

    if verbose:
        click.echo("Generating ...")
    try:
        latents = tf.random.normal((1, latent_size))

        if multi_resolution:
            images = []
            model_paths = glob.glob(os.path.join(model, "generator_res*"))
            resolutions = []

            for model_path in model_paths:
                generator = tf.saved_model.load(model_path)
                generate = generator.signatures["serving_default"]
                img = generate(latents)["generated"]
                img = np.squeeze(img)
                images.append(img)
                resolutions.append(os.path.splitext(model_path)[0].split("_")[-1])

        else:
            output_resolution = int(output_shape[0])
            model = os.path.join(
                model, "generator_res_{}".format(output_resolution), "weights"
            )
            generator = tf.saved_model.load(model)
            generate = generator.signatures["serving_default"]
            img = generate(latents)["generated"]
            img = np.squeeze(img)

    except Exception:
        click.echo(click.style("ERROR: generation failed. See error trace.", fg="red"))
        raise

    if verbose:
        click.echo("Saving ...")
    if multi_resolution:
        for img, resolution in zip(images, resolutions):
            img = nib.Nifti1Image(img.astype(np.uint8), np.eye(4))
            # Add resolution to the outfile as an id
            nib.save(
                img,
                "{0}_res_{3}.{1}.{2}".format(*outfile.split(".") + [int(resolution)]),
            )
    else:
        img = nib.Nifti1Image(img.astype(np.uint8), np.eye(4))
        nib.save(img, outfile)

    if verbose:
        click.echo("Output saved to {}".format(outfile))


@cli.command()
def save():
    """Save a model to SavedModel type."""
    click.echo(
        "Not implemented yet. In the future, this command will be used for saving."
    )
    sys.exit(-2)


@cli.command()
def evaluate():
    """Evaluate a model's predictions against known labels."""
    click.echo(
        "Not implemented yet. In the future, this command will be used for evaluation."
    )
    sys.exit(-2)


@cli.command()
def info():
    """Return information about this system."""
    uname = platform.uname()
    s = f"""\
Python:
 Version: {platform.python_version()}
 Implementation: {platform.python_implementation()}
 64-bit: {sys.maxsize > 2**32}
 Packages:
  Nobrainer: {__version__}
  Nibabel: {nib.__version__}
  Numpy: {np.__version__}
  TensorFlow: {tf.__version__}
   GPU support: {tf.test.is_built_with_gpu_support()}
   GPU available: {bool(tf.config.list_physical_devices('GPU'))}
  TensorFlow-Probability: {tfp.__version__}

System:
 OSType: {uname.system}
 Release: {uname.release}
 Version: {uname.version}
 Architecture: {uname.machine}

Timestamp: {datetime.datetime.utcnow().strftime('%Y/%m/%d %T')}"""
    click.echo(s)


# For debugging only.
if __name__ == "__main__":
    cli()



from .dropout import BernoulliDropout, ConcreteDropout, GaussianDropout
from .padding import ZeroPadding3DChannels

__all__ = [
    "BernoulliDropout",
    "ConcreteDropout",
    "GaussianDropout",
    "ZeroPadding3DChannels",
]



import numpy as np
from numpy.testing import assert_array_equal
import pytest
import tensorflow as tf

from .. import volume


def test_binarize():
    x = [
        0.49671415,
        -0.1382643,
        0.64768854,
        1.52302986,
        -0.23415337,
        -0.23413696,
        1.57921282,
        0.76743473,
    ]
    x = np.asarray(x, dtype="float64")
    expected = np.array([True, False, True, True, False, False, True, True])
    result = volume.binarize(x)
    assert_array_equal(expected, result)
    assert result.dtype == tf.float64
    result = volume.binarize(x.astype(np.float32))
    assert_array_equal(expected, result)
    assert result.dtype == tf.float32

    x = np.asarray([-2, 0, 2, 0, 2, -2, -1, 1], dtype=np.int32)
    expected = np.array([False, False, True, False, True, False, False, True])
    result = volume.binarize(x)
    assert_array_equal(expected, result)
    assert result.dtype == tf.int32
    result = volume.binarize(x.astype(np.int64))
    assert_array_equal(expected, result)
    assert result.dtype == tf.int64


@pytest.mark.parametrize("replace_func", [volume.replace, volume.replace_in_numpy])
def test_replace(replace_func):
    data = np.arange(5)
    mapping = {0: 10, 1: 20, 2: 30, 3: 40, 4: 30}
    output = replace_func(data, mapping)
    assert_array_equal(output, [10, 20, 30, 40, 30])

    # Test that overlapping keys and values gives correct result.
    data = np.arange(5)
    mapping = {0: 1, 1: 2, 2: 3, 3: 4}
    output = replace_func(data, mapping)
    assert_array_equal(output, [1, 2, 3, 4, 0])

    data = np.arange(8).reshape(2, 2, 2)
    mapping = {0: 100, 100: 10, 10: 5, 3: 5}
    outputs = replace_func(data, mapping, zero=False)
    expected = data.copy()
    expected[0, 0, 0] = 100
    expected[0, 1, 1] = 5
    assert_array_equal(outputs, expected)

    # Zero values not in mapping values.
    outputs = replace_func(data, mapping, zero=True)
    expected = np.zeros_like(data)
    expected[0, 0, 0] = 100
    expected[0, 1, 1] = 5
    assert_array_equal(outputs, expected)


@pytest.mark.parametrize("std_func", [volume.standardize, volume.standardize_numpy])
def test_standardize(std_func):
    x = np.random.randn(10, 10, 10).astype(np.float32)
    outputs = np.array(std_func(x))
    assert np.allclose(outputs.mean(), 0, atol=1e-07)
    assert np.allclose(outputs.std(), 1, atol=1e-07)

    if std_func == volume.standardize:
        x = np.random.randn(10, 10, 10).astype(np.float64)
        outputs = np.array(std_func(x))
        assert outputs.dtype == np.float32


@pytest.mark.parametrize("norm_func", [volume.normalize, volume.normalize_numpy])
def test_normalize(norm_func):
    x = np.random.randn(10, 10, 10).astype(np.float32)
    outputs = np.array(norm_func(x))
    assert np.allclose(outputs.min(), 0, atol=1e-07)
    assert np.allclose(outputs.max(), 1, atol=1e-07)

    if norm_func == volume.normalize:
        x = np.random.randn(10, 10, 10).astype(np.float64)
        outputs = np.array(norm_func(x))
        assert outputs.dtype == np.float32


def _stack_channels(_in):
    return np.stack([_in, 2 * _in, 3 * _in], axis=-1)


@pytest.mark.parametrize("multichannel", [True, False])
@pytest.mark.parametrize("to_blocks_func", [volume.to_blocks, volume.to_blocks_numpy])
def test_to_blocks(multichannel, to_blocks_func):
    x = np.arange(8).reshape(2, 2, 2)
    block_shape = (1, 1, 1)
    if multichannel:
        x = _stack_channels(x)
        block_shape = (1, 1, 1, 3)
    outputs = np.array(to_blocks_func(x, block_shape))
    expected = np.array(
        [[[[0]]], [[[1]]], [[[2]]], [[[3]]], [[[4]]], [[[5]]], [[[6]]], [[[7]]]]
    )
    if multichannel:
        expected = _stack_channels(expected)
    assert_array_equal(outputs, expected)

    block_shape = 2
    if multichannel:
        block_shape = (2, 2, 2, 3)
    outputs = np.array(to_blocks_func(x, block_shape))
    assert_array_equal(outputs, x[None])

    block_shape = (3, 3, 3)
    if multichannel:
        block_shape = (3, 3, 3, 3)
    with pytest.raises((tf.errors.InvalidArgumentError, ValueError)):
        to_blocks_func(x, block_shape)

    block_shape = (3, 3)
    with pytest.raises(ValueError):
        to_blocks_func(x, block_shape)


@pytest.mark.parametrize("multichannel", [True, False])
@pytest.mark.parametrize(
    "from_blocks_func", [volume.from_blocks, volume.from_blocks_numpy]
)
def test_from_blocks(multichannel, from_blocks_func):
    x = np.arange(64).reshape(4, 4, 4)
    block_shape = (2, 2, 2)
    if multichannel:
        x = _stack_channels(x)
        block_shape = (2, 2, 2, 3)

    outputs = from_blocks_func(volume.to_blocks(x, block_shape), x.shape)
    assert_array_equal(outputs, x)

    with pytest.raises(ValueError):
        x = np.arange(80).reshape(10, 2, 2, 2)
        outputs = from_blocks_func(x, (4, 4, 4))


def test_blocks_numpy_value_errors():
    with pytest.raises(ValueError):
        x = np.random.rand(4, 4)
        output_shape = (4, 4, 4)
        volume.to_blocks_numpy(x, output_shape)

    with pytest.raises(ValueError):
        x = np.random.rand(4, 4, 4)
        output_shape = (4, 4, 4)
        volume.from_blocks_numpy(x, output_shape)

    with pytest.raises(ValueError):
        x = np.random.rand(4, 4, 4, 4)
        output_shape = (4, 4)
        volume.from_blocks_numpy(x, output_shape)



# TO DO def DivisiblePad(x,y= None, trans_xy= False, k):
import numpy as np
import tensorflow as tf


def addGaussianNoise(x, y=None, trans_xy=False, noise_mean=0.0, noise_std=0.1):
    """Add Gaussian noise to input and label.

    Usage:
    ```python
    >>> x = [[[1., 1., 1.]]]
    >>> x_out = intensity_transforms.addGaussianNoise(x,
                                  noise_mean=0.0, noise_std=1)
    >>> x_out
        <tf.Tensor: shape=(1, 1, 3), dtype=float32,
        numpy=array([[[0.82689023, 1.9072294 , 1.9717102 ]]], dtype=float32)>
    ```

    Parameters
    ----------
        x: input is a tensor or numpy to have rank 3,
        y: label is a tensor or numpy to have rank 3,
        noise_mean: int, mean of Gaussian kernel. Default = 0.0;
        noise_std: int, standard deviation of Gaussian kernel. Default=0.1;
        trans_xy: Boolean, transforms both x and y. If set True, function
        will require both x,y.

    Returns
    ----------
        Input and/or label tensor with added Gaussian noise.
    """
    if ~tf.is_tensor(x):
        x = tf.convert_to_tensor(x)
    x = tf.cast(x, tf.float32)
    noise = tf.random.normal(x.shape, noise_mean, noise_std, dtype=x.dtype)
    if trans_xy:
        if y is None:
            raise ValueError("`LabelMap' should be assigned")
        if ~tf.is_tensor(y):
            y = tf.convert_to_tensor(y)
        if len(y.shape) != 3:
            raise ValueError("`LabelMap` must be equal or higher than rank 2")

        y = tf.cast(y, tf.float32)
        return tf.math.add(x, noise), tf.math.add(y, noise)
    if y is None:
        return tf.math.add(x, noise)
    return tf.math.add(x, noise), y


def minmaxIntensityScaling(x, y=None, trans_xy=False):
    """Apply intensity scaling [0-1] to input and label.

    Usage:
    ```python
    >>> x = [[[0., 2., 1.]]]
    >>> x_out = intensity_transforms.minmaxIntensityScaling(x)
    >>> x_out
    <tf.Tensor: shape=(1, 1, 3), dtype=float32,
    numpy=array([[[0., 1. , 0.5]]], dtype=float32)>
    ```

    Parameters
    ----------
        x: input is a tensor or numpy to have rank 3,
        y: label is a tensor or numpy to have rank 3,
        trans_xy: Boolean, transforms both x and y. If set True, function
        will require both x,y.

    Returns
    ----------
        Input and/or label tensor with scaled intensity.
    """
    if ~tf.is_tensor(x):
        x = tf.convert_to_tensor(x)
    x = tf.cast(x, tf.float32)
    ep = tf.cast(
        tf.convert_to_tensor(1e-8 * np.ones(x.shape).astype(np.float32)), tf.float32
    )
    xmin = tf.cast(tf.reduce_min(x), tf.float32)
    xmax = tf.cast(tf.reduce_max(x), tf.float32)
    x = tf.divide(tf.subtract(x, xmin), tf.add(tf.subtract(xmax, xmin), ep))
    if trans_xy:
        if y is None:
            raise ValueError("`LabelMap' should be assigned")
        if len(y.shape) != 3:
            raise ValueError("`LabelMap` must be equal or higher than rank 2")
        if ~tf.is_tensor(y):
            y = tf.convert_to_tensor(y)
        y = tf.cast(y, tf.float32)
        ymin = tf.cast(tf.reduce_min(y), tf.float32)
        ymax = tf.cast(tf.reduce_max(y), tf.float32)
        y = tf.divide(tf.subtract(y, ymin), tf.add(tf.subtract(ymax, ymin), ep))
    if y is None:
        return x
    return x, y


def customIntensityScaling(x, y=None, trans_xy=False, scale_x=[0.0, 1.0], scale_y=None):
    """Apply custom intensity scaling to input and label.

    Usage:
    ```python
    >>> x = [[[2., 2., 1.]]]
    >>> y = [[[1., 0., 1.]]]
    >>> x_out, y_out = intensity_transforms.customIntensityScaling(
    x, y, trans_xy=True, scale_x=[0, 4], scale_y=[0, 3])
    >>> x_out
    <tf.Tensor: shape=(1, 1, 3), dtype=float32,
    numpy=array([[[4., 4., 0.]]], dtype=float32)>
    >>> y_out
    <tf.Tensor: shape=(1, 1, 3), dtype=float32,
    numpy=array([[[3., 0., 3.]]], dtype=float32)>
    ```

    Parameters
    ----------
        x: input is a tensor or numpy to have rank 3,
        y: label is a tensor or numpy to have rank 3,
        trans_xy: Boolean, transforms both x and y (Default: False).
           If set True, function will require both x,y.
        scale_x: [minimum(int), maximum(int)]
        scale_y: [minimum(int), maximum(int)]

    Returns
    ----------
        Input and/or label tensor with custom scaled Intensity.
    """
    x_norm, y_norm = minmaxIntensityScaling(x, y, trans_xy)
    minx = tf.cast(
        tf.convert_to_tensor(scale_x[0] * np.ones(x_norm.shape).astype(np.float32)),
        tf.float32,
    )
    maxx = tf.cast(
        tf.convert_to_tensor(scale_x[1] * np.ones(x_norm.shape).astype(np.float32)),
        tf.float32,
    )

    diff_x = tf.subtract(maxx, minx)
    x = tf.add(tf.multiply(x_norm, diff_x), minx)
    if trans_xy:
        if y is None:
            raise ValueError("`LabelMap' should be assigned")
        if scale_y is None:
            raise ValueError("LabelMap scaling arguments as: scale_Y=[a,b]")
        y = tf.cast(y, tf.float32)
        miny = tf.cast(
            tf.convert_to_tensor(scale_y[0] * np.ones(y_norm.shape).astype(np.float32)),
            tf.float32,
        )
        maxy = tf.cast(
            tf.convert_to_tensor(scale_y[1] * np.ones(y_norm.shape).astype(np.float32)),
            tf.float32,
        )
        diff_y = tf.subtract(maxy, miny)
        y = tf.add(tf.multiply(y_norm, diff_y), miny)
    if y is None:
        return x
    return x, y


def intensityMasking(x, mask_x, y=None, trans_xy=False, mask_y=None):
    """Masking the intensity values in input and label.

    Usage:
    ```python
    >>> mask_x = np.array([[[0, 0, 0], [0, 1, 0], [0, 0, 0]]])
    >>> x = np.array([[[1, 1, 1], [2, 2, 2], [3, 3, 3]]])
    >>> x_out = intensity_transforms.intensityMasking(x,
                mask_x=mask_x)
    >>> x_out
    (<tf.Tensor: shape=(2, 3, 3), dtype=float32, numpy=
     array([[[0., 0., 0.],
             [0., 2., 0.],
             [0., 0., 0.]], dtype=float32)>, None)
    ```

    Parameters
    ----------
        x: input is a tensor or numpy to have rank 3,
        y: label is a tensor or numpy to have rank 3,
        mask_x: mask tensor or numpy array of same shape as x
        trans_xy: Boolean, transforms both x and y (Default: False).
           If set True, function will require both x,y.

    Returns
    ----------
        Masked input and/or label tensor.
    """
    if ~tf.is_tensor(x):
        x = tf.convert_to_tensor(x)
    x = tf.cast(x, tf.float32)
    if ~tf.is_tensor(mask_x):
        mask_x = tf.convert_to_tensor(mask_x)
    mask_x = tf.cast(mask_x, tf.float32)
    if mask_x.shape[0] != x.shape[0] and mask_x.shape[1] != x.shape[1]:
        raise ValueError("Masks shape should be same as Input")
    x = tf.multiply(x, mask_x)
    if trans_xy:
        if y is None:
            raise ValueError("`LabelMap' should be assigned")
        if mask_y is None:
            raise ValueError("Label Mask should not be none")
        if ~tf.is_tensor(y) and ~tf.is_tensor(mask_y):
            y = tf.convert_to_tensor(y)
            mask_y = tf.convert_to_tensor(mask_y)
        y = tf.cast(y, tf.float32)
        mask_x = tf.cast(mask_x, tf.float32)
        if mask_y.shape[0] != y.shape[0] and mask_x.shape[1] != x.shape[1]:
            raise ValueError("Label Masks shape should be same as Label")
        return x, tf.multiply(y, mask_y)
    if y is None:
        return x
    return x, y


def contrastAdjust(x, y=None, trans_xy=False, gamma=1.0):
    """Apply contrast adjustment to input and label.

    Usage:
    ```python
    >>> gamma = 1.5
    >>> epsilon = 1e-7
    >>> x = np.array([[[1, 1, 1], [2, 2, 2], [3, 3, 3]]])
    >>> x_out
    (<tf.Tensor: shape=(1, 3, 3), dtype=float32, numpy=
     array([[[1.       , 1.       , 1.       ],
             [1.7071067, 1.7071067, 1.7071067],
             [3.       , 3.       , 3.       ]]], dtype=float32)>, None)
    ```

    Parameters
    ----------
        x: input is a tensor or numpy to have rank 3,
        y: label is a tensor or numpy to have rank 3,
        gamma: int, a contrast adjustment constant
        trans_xy: Boolean, transforms both x and y (Default: False).
           If set True, function will require both x,y.

    Returns
    ----------
        Input and/or label tensor with adjusted contrast.
    """
    if ~tf.is_tensor(x):
        x = tf.convert_to_tensor(x)
    x = tf.cast(x, tf.float32)
    ep = tf.cast(
        tf.convert_to_tensor(1e-7 * np.ones(x.shape).astype(np.float32)), tf.float32
    )
    gamma = tf.cast(
        tf.convert_to_tensor(gamma * np.ones(x.shape).astype(np.float32)), tf.float32
    )
    xmin = tf.cast(tf.reduce_min(x), tf.float32)
    xmax = tf.cast(tf.reduce_max(x), tf.float32)
    x_range = tf.subtract(xmax, xmin)
    x = tf.pow(tf.divide(tf.subtract(x, xmin), tf.add(x_range, ep)), gamma)
    x = tf.add(tf.multiply(x, x_range), xmin)
    if trans_xy:
        if y is None:
            raise ValueError("`LabelMap' should be assigned")
        if len(y.shape) != 3:
            raise ValueError("`LabelMap` must be equal or higher than rank 2")
        if ~tf.is_tensor(y):
            y = tf.convert_to_tensor(y)
        y = tf.cast(y, tf.float32)
        ymin = tf.cast(tf.reduce_min(y), tf.float32)
        ymax = tf.cast(tf.reduce_max(y), tf.float32)
        y_range = tf.subtract(ymax, ymin)
        y = tf.pow(tf.divide(tf.subtract(y, ymin), tf.add(y_range, ep)), gamma)
        y = tf.add(tf.multiply(y, y_range), ymin)
        return x, y
    if y is None:
        return x
    return x, y



"""Methods for creating `tf.data.Dataset` objects."""

import math
import os
from pathlib import Path

import fsspec
import nibabel as nb
import numpy as np
import tensorflow as tf

from .io import _is_gzipped, verify_features_labels
from .tfrecord import _labels_all_scalar, parse_example_fn, write
from .volume import binarize, replace, to_blocks

AUTOTUNE = tf.data.experimental.AUTOTUNE


def write_multi_resolution(
    paths,
    tfrecdir=Path(os.getcwd()) / "data",
    resolutions=None,
    shard_size=3,
    n_processes=1,
):
    """Create a multiresolution dataset.

    This returns a dictionary of information that is used by the generative model
    processing class.

    TODO: This function needs to be aligned with the Dataset class
    """
    resolutions = resolutions or [8, 16, 32, 64, 128, 256]
    tfrecdir = Path(tfrecdir)
    tfrecdir.mkdir(exist_ok=True, parents=True)
    template = tfrecdir / "data-train_shard-{shard:03d}.tfrec"

    write(
        features_labels=paths,
        filename_template=str(template),
        examples_per_shard=shard_size,  # change for larger dataset
        multi_resolution=True,
        resolutions=resolutions,
        processes=n_processes,
    )

    datasets = {}
    for resolution in resolutions:
        datasets[resolution] = dict(
            file_pattern=str(tfrecdir / f"*res-{resolution:03d}.tfrec"),
            batch_size=1,
            normalizer=None,
        )
    return datasets


class Dataset:
    """Datasets for training, and validation"""

    def __init__(
        self,
        dataset,
        n_volumes,
        volume_shape,
        n_classes,
    ):
        self.dataset = dataset
        self.n_volumes = n_volumes
        self.volume_shape = volume_shape
        self.n_classes = n_classes

    @classmethod
    def from_tfrecords(
        cls,
        file_pattern=None,
        n_volumes=None,
        volume_shape=None,
        block_shape=None,
        scalar_labels=False,
        n_classes=1,
        tf_dataset_options=None,
        num_parallel_calls=1,
    ):
        """Function to retrieve a saved tf record as a nobrainer Dataset

        file_pattern: str, the path from which TFRecord files should be read.
        num_parallel_calls: int, number of processes to use for multiprocessing. If
            None, will use all available processes.
        """

        fs, _, _ = fsspec.get_fs_token_paths(file_pattern)
        files = fs.glob(file_pattern)
        if not files:
            raise ValueError("no files found for pattern '{}'".format(file_pattern))

        # Create dataset of all TFRecord files. After this point, the dataset will have
        # two value per iteration: (feature, label).
        compressed = _is_gzipped(files[0], filesys=fs)
        dataset = tf.data.Dataset.list_files(file_pattern, shuffle=False)

        if tf_dataset_options:
            dataset = dataset.with_options(tf_dataset_options)

        # Read each of these files as a TFRecordDataset.
        # Assume all files have same compression type as the first file.
        compression_type = "GZIP" if compressed else None
        cycle_length = 1 if num_parallel_calls is None else num_parallel_calls
        parse_fn = parse_example_fn(
            volume_shape=volume_shape, scalar_labels=scalar_labels
        )

        # Determine examples_per_shard from the first TFRecord shard
        # Then set block_length to equal the number of examples per shard
        # so that the interleave method does not inadvertently shuffle data.
        first_shard = (
            dataset.take(1)
            .flat_map(
                lambda x: tf.data.TFRecordDataset(x, compression_type=compression_type)
            )
            .map(map_func=parse_fn, num_parallel_calls=num_parallel_calls)
        )
        block_length = len([0 for _ in first_shard])

        if not n_volumes:
            n_volumes = block_length * len(files)

        dataset = dataset.interleave(
            map_func=lambda x: tf.data.TFRecordDataset(
                x, compression_type=compression_type
            ),
            cycle_length=cycle_length,
            block_length=block_length,
            num_parallel_calls=num_parallel_calls,
        )
        dataset = dataset.map(map_func=parse_fn, num_parallel_calls=num_parallel_calls)
        ds_obj = cls(dataset, n_volumes, volume_shape, n_classes)

        if block_shape:
            ds_obj.block(block_shape)
        if not scalar_labels:
            ds_obj.map_labels()
        # TODO automatically determine batch size
        ds_obj.batch(1)

        return ds_obj

    @classmethod
    def from_files(
        cls,
        filepaths,
        check_shape=True,
        check_labels_int=False,
        check_labels_gte_zero=False,
        out_tfrec_dir=Path(os.getcwd()) / "data",
        shard_size=300,
        num_parallel_calls=1,
        eval_size=0.1,
        n_classes=1,
        block_shape=None,
        tf_dataset_options=None,
    ):
        """Create Nobrainer datasets from data
        filepaths: List(str), list of paths to individual input data files.
        check_shape: boolean, if true, validate that the shape of both volumes is
            equal to 'volume_shape'.
        check_labels_int: boolean, if true, validate that every labels volume is an
            integer type or can be safely converted to an integer type.
        check_labels_gte_zero: boolean, if true, validate that every labels volume
            has values greater than or equal to zero.
        out_tfrec_dir: str, the directory to which TFRecord files should be written.
        shard_size: int, number of pairs of `(feature, label)` per TFRecord file.
        num_parallel_calls: int, number of processes to use for multiprocessing. If
            None, will use all available processes.
        eval_size: float, proportion of the input files to reserve for validation.
        n_classes: int, number of output classes
        """
        n_eval = np.ceil(len(filepaths) * eval_size).astype(int)
        n_train = len(filepaths) - n_eval

        verify_result = verify_features_labels(
            filepaths,
            check_shape=check_shape,
            check_labels_int=check_labels_int,
            check_labels_gte_zero=check_labels_gte_zero,
        )
        if len(verify_result) != 0:
            raise ValueError(
                "Provided filepaths did not pass validation. Please "
                "check that they have the same shape, and the "
                "targets have appropriate labels"
            )

        Path(out_tfrec_dir).mkdir(exist_ok=True, parents=True)
        template = str(Path(out_tfrec_dir) / "data-{intent}")
        volume_shape = nb.load(filepaths[0][0]).shape
        write(
            features_labels=filepaths[:n_train],
            filename_template=template.format(intent="train")
            + "_shard-{shard:03d}.tfrec",
            examples_per_shard=shard_size,
            processes=num_parallel_calls,
        )
        if n_eval > 0:
            write(
                features_labels=filepaths[n_train:],
                filename_template=template.format(intent="eval")
                + "_shard-{shard:03d}.tfrec",
                examples_per_shard=shard_size,
                processes=num_parallel_calls,
            )
        labels = (y for _, y in filepaths)
        scalar_labels = _labels_all_scalar(labels)
        # replace shard formatting code with * for globbing
        template_train = template.format(intent="train_*.tfrec")
        ds_train = cls.from_tfrecords(
            template_train,
            n_train,
            volume_shape,
            scalar_labels=scalar_labels,
            n_classes=n_classes,
            block_shape=block_shape,
            tf_dataset_options=tf_dataset_options,
            num_parallel_calls=num_parallel_calls,
        )
        ds_eval = None
        if n_eval > 0:
            template_eval = template.format(intent="eval_*.tfrec")
            ds_eval = cls.from_tfrecords(
                template_eval,
                n_eval,
                volume_shape,
                scalar_labels=scalar_labels,
                n_classes=n_classes,
                block_shape=block_shape,
                tf_dataset_options=tf_dataset_options,
                num_parallel_calls=num_parallel_calls,
            )
        return ds_train, ds_eval

    @property
    def batch_size(self):
        return self.dataset.element_spec[0].shape[0]

    @property
    def block_shape(self):
        return tuple(self.dataset.element_spec[0].shape[1:4].as_list())

    @property
    def scalar_labels(self):
        return _labels_all_scalar([y for _, y in self.dataset.as_numpy_iterator()])

    def get_steps_per_epoch(self):
        def get_n(a, k):
            return (a - k) / k + 1

        n_blocks = tuple(
            get_n(aa, kk) for aa, kk in zip(self.volume_shape, self.block_shape)
        )

        for n in n_blocks:
            if not n.is_integer() or n < 1:
                raise ValueError(
                    "cannot create non-overlapping blocks with the given parameters."
                )
        n_blocks_per_volume = np.prod(n_blocks).astype(int)

        steps = n_blocks_per_volume * self.n_volumes / self.batch_size
        steps = math.ceil(steps)
        return steps

    def map(self, func, num_parallel_calls=AUTOTUNE):
        self.dataset = self.dataset.map(func, num_parallel_calls=num_parallel_calls)
        return self

    def normalize(self, normalizer):
        return self.map(lambda x, y: (normalizer(x), y))

    def augment(self, augment_steps, num_parallel_calls=AUTOTUNE):
        batch_size = None
        if len(self.dataset.element_spec[0].shape) > 4:
            batch_size = self.batch_size
            self.dataset = self.dataset.unbatch()

        for transform, kwargs in augment_steps:
            self.map(
                lambda x, y: tf.cond(
                    tf.random.uniform((1,)) > 0.5,
                    true_fn=lambda: transform(x, y, **kwargs),
                    false_fn=lambda: (x, y),
                ),
                num_parallel_calls=num_parallel_calls,
            )

        if batch_size:
            self.batch(batch_size)

        return self

    def block(self, block_shape, num_parallel_calls=AUTOTUNE):
        if not self.scalar_labels:
            self.map(
                lambda x, y: (to_blocks(x, block_shape), to_blocks(y, block_shape)),
                num_parallel_calls=num_parallel_calls,
            )
        else:

            def _f(x, y):
                x = to_blocks(x, block_shape)
                n_blocks = x.shape[0]
                y = tf.repeat(y, n_blocks)
                return (x, y)

            self.map(_f, num_parallel_calls=num_parallel_calls)
        # This step is necessary because separating into blocks adds a dimension.
        self.dataset = self.dataset.unbatch()
        return self

    def map_labels(self, label_mapping=None):
        if self.n_classes < 1:
            raise ValueError("n_classes must be > 0.")

        if label_mapping is not None:
            self.map(lambda x, y: (x, replace(y, label_mapping=label_mapping)))

        if self.n_classes == 1:
            self.map(lambda x, y: (x, tf.expand_dims(binarize(y), -1)))
        elif self.n_classes == 2:
            self.map(lambda x, y: (x, tf.one_hot(binarize(y), self.n_classes)))
        elif self.n_classes > 2:
            self.map(lambda x, y: (x, tf.one_hot(y, self.n_classes)))

        return self

    def batch(self, batch_size):
        # If volume_shape is only three dims, add grayscale channel to features.
        # Otherwise, assume that the channels are already in the features.
        if len(self.dataset.element_spec[0].shape) == 3:
            self.map(lambda x, y: (tf.expand_dims(x, -1), y))
        elif len(self.dataset.element_spec[0].shape) > 4:
            self.dataset = self.dataset.unbatch()

        # Prefetch data to overlap data production with data consumption. The
        # TensorFlow documentation suggests prefetching `batch_size` elements.
        self.dataset = self.dataset.prefetch(buffer_size=batch_size)

        # Batch the dataset, so each iteration gives `batch_size` elements. We drop
        # the remainder so that when training on multiple GPUs, the batch will
        # always be evenly divisible by the number of GPUs. Otherwise, the last
        # batch might have fewer than `batch_size` elements and will cause errors.
        self.dataset = self.dataset.batch(batch_size=batch_size, drop_remainder=True)

        return self

    def shuffle(self, shuffle_buffer_size):
        # Optionally shuffle. We also optionally shuffle the list of files.
        # The TensorFlow recommend shuffling and then repeating.
        self.dataset = self.dataset.shuffle(buffer_size=shuffle_buffer_size)
        return self

    def repeat(self, n_repeats):
        # Repeat the dataset for n_epochs. If n_epochs is None, then repeat
        # indefinitely. If n_epochs is 1, then the dataset will only be iterated
        # through once.
        self.dataset = self.dataset.repeat(n_repeats)
        return self



import csv

import nibabel as nib
import numpy as np
from numpy.testing import assert_array_equal
import pytest
from scipy.stats import entropy

from ..utils import StreamingStats


@pytest.fixture(scope="session")
def csv_of_volumes(tmpdir_factory):
    """Create random Nifti volumes for use in testing, and return filepath to
    CSV, which contains rows of filepaths to `(features, labels)`.
    """
    savedir = tmpdir_factory.mktemp("data")
    volume_shape = (8, 8, 8)
    n_volumes = 100

    features = np.random.rand(n_volumes, *volume_shape).astype(np.float32) * 10
    labels = np.random.randint(0, 1, size=(n_volumes, *volume_shape))
    labels = labels.astype(np.int32)
    affine = np.eye(4)
    list_of_filepaths = []

    for idx in range(n_volumes):
        fpf = str(savedir.join("{}f.nii.gz")).format(idx)
        fpl = str(savedir.join("{}l.nii.gz")).format(idx)

        nib.save(nib.Nifti1Image(features[idx], affine), fpf)
        nib.save(nib.Nifti1Image(labels[idx], affine), fpl)
        list_of_filepaths.append((fpf, fpl))

    filepath = savedir.join("features_labels.csv")
    with open(filepath, mode="w", newline="") as f:
        writer = csv.writer(f, delimiter=",")
        writer.writerows(list_of_filepaths)

    return str(filepath)


def test_stream_stat():
    s1 = np.array([[0.5, 0.1, 0.4]])
    s2 = np.array([[0.5, 0.2, 0.3]])
    s3 = np.array([[0.4, 0.2, 0.4]])

    st = np.concatenate((s1, s2, s3), axis=0)

    s = StreamingStats()
    s.update(s1).update(s2).update(s3)

    assert_array_equal(s.mean(), np.mean(st, axis=0))
    assert_array_equal(s.var(), np.var(st, axis=0))
    assert_array_equal(s.std(), np.std(st, axis=0))
    assert_array_equal(np.sum(s.entropy()), entropy(np.mean(st, axis=0), axis=0))



"""Model definition for ProgressiveAE.
"""
import tensorflow as tf
from tensorflow.keras import layers, models


def progressiveae(
    latent_size,
    label_size=0,
    num_channels=1,
    dimensionality=3,
    e_fmap_base=2048,
    d_fmap_base=2048,
    e_fmap_max=512,
    d_fmap_max=256,
):
    """Instantiate ProgressiveAE Architecture.

    Parameters
    ----------
    latent_size: int, size of the latent space to use for generating images
    label_size: int, number of labels for conditional image synthesis (WIP)
    num_channels: int, number of channels in the generated images
    dimensionality: int, one of [2, 3], number of dimensions in the image
    e_fmap_base: int, parameter to determine width of encoder
    d_fmap_base: int, parameter to determine width of decoder
    e_fmap_max: int, parameter to determine width of encoder
    d_fmap_max: int, parameter to determine width of decoder

    Returns
    -------
    Encoder and Decoder
    """
    encoder = Encoder(
        latent_size,
        label_size=label_size,
        num_channels=num_channels,
        dimensionality=dimensionality,
        fmap_base=e_fmap_base,
        fmap_max=e_fmap_max,
    )

    decoder = Decoder(
        latent_size,
        label_size=label_size,
        num_channels=num_channels,
        dimensionality=dimensionality,
        fmap_base=d_fmap_base,
        fmap_max=d_fmap_max,
    )

    return encoder, decoder


class Encoder(tf.keras.Model):
    def __init__(
        self,
        latent_size,
        label_size=0,
        num_channels=1,
        fmap_base=2048,
        fmap_max=512,
        dimensionality=3,
    ):
        super(Encoder, self).__init__()

        self.latent_size = latent_size
        self.label_size = label_size

        self.fmap_base = fmap_base
        self.fmap_max = fmap_max

        self.num_channels = num_channels
        self.dimensionality = dimensionality

        self.current_resolution = 2
        self.current_width = 2**self.current_resolution

        self.Conv = getattr(layers, "Conv{}D".format(self.dimensionality))
        self.Upsampling = getattr(layers, "UpSampling{}D".format(self.dimensionality))
        self.AveragePooling = getattr(
            layers, "AveragePooling{}D".format(self.dimensionality)
        )

        self.highest_resolution_block = self.make_Eblock(
            (self._nf(self.current_resolution)),
            name=("e_block_{}".format(self.current_resolution)),
        )
        self.resolution_blocks = []

        self.Base_Dense = layers.Dense(self.latent_size)
        self.Head_Conv = self.Conv(
            (self._nf(self.current_resolution)),
            kernel_size=1,
            padding="same",
        )
        self.Base_Conv = self.Conv(
            (self._nf(self.current_resolution - 1)),
            kernel_size=1,
            padding="same",
        )

        images_shape = (
            (None,)
            + (int(2.0**self.current_resolution),) * self.dimensionality
            + (self.num_channels,)
        )
        alpha_shape = (1,)

        self.build([images_shape, alpha_shape])

    def _pixel_norm(self, epsilon=1e-08):
        """
        Pixelwise normalization
        """
        return layers.Lambda(
            lambda x: x
            * tf.math.rsqrt(
                tf.reduce_mean((tf.square(x)), axis=(-1), keepdims=True) + epsilon
            )
        )

    def _weighted_sum(self):
        """
        Weighted interpolation helper for fading in new layers as its progressively added
        """
        return layers.Lambda(
            lambda inputs: (1 - inputs[2]) * inputs[0] + (inputs[2]) * inputs[1]
        )

    def _nf(self, stage):
        """
        Computes number of filters for a conv layer
        """
        return min(int(self.fmap_base / (2.0 ** (stage))), self.fmap_max)

    def update_res(self):
        """
        Updates the current resolution of the model
        """
        self.current_resolution += 1
        self.current_width = 2**self.current_resolution

    def make_Ehead(self, x):
        """
        Creates the end of the encoder. The output node gives the latent representation.
        """
        x = layers.Flatten()(x)
        x = self.Base_Dense(x)

        return x

    def make_Eblock(self, nf, kernel_size=4, name="eblock", input_shape=None):
        """
        Creates an encoder block
        """
        block_layers = []
        block_layers.append(
            self.Conv(nf, kernel_size=kernel_size, strides=2, padding="same")
        )
        block_layers.append(layers.Activation(tf.nn.leaky_relu))
        return models.Sequential(block_layers, name=name)

    def make_Ebase(self, x, y, alpha):
        """
        Creates an encoder base. Performs the fading using the alpha factor
        """

        x = self.AveragePooling()(x)
        x = self.Base_Conv(x)
        x = self._weighted_sum()([x, y, alpha])
        return x

    def add_resolution(self):
        """
        Adds a resolution step to the trainable encoder by creating a
        encoder block and inserting at the start of the existing encoder
        to increase the resolution of the images encoded by a factor of 2.
        The new layers are faded in and controlled by the alpha parameter.
        """

        self.update_res()

        self.resolution_blocks.append(self.highest_resolution_block)
        self.highest_resolution_block = self.make_Eblock(
            self._nf(self.current_resolution - 1),
            name="eblock{}".format(self.current_resolution),
        )

        self.Base_Conv = self.Conv(
            filters=self._nf(self.current_resolution - 1),
            kernel_size=1,
            padding="same",
        )
        self.Head_Conv = self.Conv(
            filters=self._nf(self.current_resolution), kernel_size=1, padding="same"
        )

        images_shape = (
            (None,)
            + (int(2.0**self.current_resolution),) * self.dimensionality
            + (self.num_channels,)
        )
        alpha_shape = (1,)

        self.build([images_shape, alpha_shape])

    def call(self, inputs):
        images, alpha = inputs

        x = self.Head_Conv(images)
        y = self.highest_resolution_block(x)
        x = self.make_Ebase(images, y, alpha)

        for e_block in self.resolution_blocks[::-1]:
            x = e_block(x)

        return self.make_Ehead(x)


class Decoder(tf.keras.Model):
    def __init__(
        self,
        latent_size,
        label_size=0,
        num_channels=1,
        fmap_base=2048,
        fmap_max=256,
        dimensionality=3,
    ):
        super(Decoder, self).__init__()

        self.latent_size = latent_size
        self.label_size = label_size

        self.fmap_base = fmap_base
        self.fmap_max = fmap_max

        self.num_channels = num_channels
        self.dimensionality = dimensionality

        self.Conv = getattr(layers, "Conv{}D".format(self.dimensionality))
        self.Upsampling = getattr(layers, "UpSampling{}D".format(self.dimensionality))

        self.current_resolution = 2
        self.current_width = 2**self.current_resolution

        self.highest_resolution_block = self.make_Dblock(
            self._nf(self.current_resolution),
            name="d_block_{}".format(self.current_resolution),
        )
        self.resolution_blocks = []

        self.Head_Conv1 = self.Conv(filters=(self.num_channels), kernel_size=1)
        self.Head_Conv2 = self.Conv(filters=(self.num_channels), kernel_size=1)
        self.Base_Dense = tf.keras.layers.Dense(
            units=self._nf(1) * 2**self.dimensionality
        )

        self.build([(None, latent_size), (1,)])

    def update_res(self):
        self.current_resolution += 1
        self.current_width = 2**self.current_resolution

    def _pixel_norm(self, epsilon=1e-08):
        """
        Pixelwise normalization
        """
        return layers.Lambda(
            lambda x: x
            * tf.math.rsqrt(
                tf.reduce_mean(tf.square(x), axis=-1, keepdims=True) + epsilon
            )
        )

    def _weighted_sum(self):
        """
        Weighted interpolation helper for fading in new layers as its progressively added
        """
        return layers.Lambda(
            lambda inputs: (1 - inputs[2]) * inputs[0] + (inputs[2]) * inputs[1]
        )

    def _nf(self, stage):
        """
        Computes number of filters for a conv layer
        """
        return min(int(self.fmap_base / (2.0 ** (stage))), self.fmap_max)

    def make_Dbase(self, latents):
        """
        Creates the base of the decoder with inputs of size latent_size
        and the alpha value to determine level of interpolation/fading
        """

        x = self._pixel_norm()(latents)
        x = self.Base_Dense(x)
        x = layers.Reshape([2] * self.dimensionality + [self._nf(1)])(x)

        return x

    def make_Dblock(self, nf, kernel_size=4, name=""):
        """
        Creates a decoder block
        """

        block_layers = []

        block_layers.append(self.Upsampling())
        block_layers.append(
            self.Conv(nf, kernel_size=kernel_size, strides=1, padding="same")
        )
        block_layers.append(layers.Activation(tf.nn.leaky_relu))

        block_layers.append(self._pixel_norm())
        block_layers.append(
            self.Conv(nf, kernel_size=kernel_size, strides=1, padding="same")
        )
        block_layers.append(layers.Activation(tf.nn.leaky_relu))

        block_layers.append(self._pixel_norm())

        return models.Sequential(block_layers, name=name)

    def make_Dhead(self, x, y, alpha):
        """
        Creates a decoder head. Performs the fading/interpolation using the alpha parameter
        """

        x = self.Upsampling()(x)
        x = self.Head_Conv1(x)

        y = self.Head_Conv2(y)

        x = self._weighted_sum()([x, y, alpha])

        x = tf.keras.layers.Activation("tanh")(x)

        return x

    def add_resolution(self):
        """
        Adds a resolution step to the trainable decoder by creating a
        decoder block and inserting at the start of the existing decoder
        to increase the resolution of the images decoder by a factor of 2.
        The new layers are faded in and controlled by the alpha parameter.
        """

        self.update_res()

        self.resolution_blocks.append(self.highest_resolution_block)
        self.highest_resolution_block = self.make_Dblock(
            self._nf(self.current_resolution),
            name="d_block_{}".format(self.current_resolution),
        )

        self.Head_Conv1 = self.Conv(filters=(self.num_channels), kernel_size=1)
        self.Head_Conv2 = self.Conv(filters=(self.num_channels), kernel_size=1)

        self.build([(None, self.latent_size), (1,)])

    def call(self, inputs):
        latents, alpha = inputs

        x = self.make_Dbase(latents)

        for d_block in self.resolution_blocks:
            x = d_block(x)

        y = self.highest_resolution_block(x)
        return self.make_Dhead(x, y, alpha)



from distutils.version import LooseVersion

import numpy as np
from numpy.testing import assert_allclose, assert_array_equal
import pytest
import scipy.spatial.distance
import tensorflow as tf

from .. import losses


def test_dice():
    x = np.zeros(4)
    y = np.zeros(4)
    out = losses.dice(x, y, axis=None).numpy()
    assert_allclose(out, 0)

    x = np.ones(4)
    y = np.ones(4)
    out = losses.dice(x, y, axis=None).numpy()
    assert_allclose(out, 0)

    x = [0.0, 0.0, 1.0, 1.0]
    y = [1.0, 1.0, 1.0, 1.0]
    out = losses.dice(x, y, axis=None).numpy()
    ref = scipy.spatial.distance.dice(x, y)
    assert_allclose(out, ref)

    x = [0.0, 0.0, 1.0, 1.0]
    y = [1.0, 1.0, 0.0, 0.0]
    out = losses.dice(x, y, axis=None).numpy()
    ref = scipy.spatial.distance.dice(x, y)
    assert_allclose(out, ref)
    assert_allclose(out, 1)

    x = np.ones((4, 32, 32, 32, 1), dtype=np.float32)
    y = x.copy()
    x[:2, :10, 10:] = 0
    y[:2, :3, 20:] = 0
    y[3:, 10:] = 0
    dices = np.empty(x.shape[0])
    for i in range(x.shape[0]):
        dices[i] = scipy.spatial.distance.dice(x[i].flatten(), y[i].flatten())
    assert_allclose(losses.dice(x, y, axis=(1, 2, 3, 4)), dices, rtol=1e-05)
    assert_allclose(losses.Dice(axis=(1, 2, 3, 4))(x, y), dices.mean(), rtol=1e-05)
    assert_allclose(losses.Dice(axis=(1, 2, 3, 4))(y, x), dices.mean(), rtol=1e-05)


def test_generalized_dice():
    shape = (8, 32, 32, 32, 16)
    x = np.zeros(shape)
    y = np.zeros(shape)
    assert_array_equal(losses.generalized_dice(x, y), np.zeros(shape[0]))

    shape = (8, 32, 32, 32, 16)
    x = np.ones(shape)
    y = np.ones(shape)
    assert_array_equal(losses.generalized_dice(x, y), np.zeros(shape[0]))

    shape = (8, 32, 32, 32, 16)
    x = np.ones(shape)
    y = np.zeros(shape)
    # Why aren't the losses exactly one? Could it be the propagation of floating
    # point inaccuracies when summing?
    assert_allclose(losses.generalized_dice(x, y), np.ones(shape[0]), atol=1e-03)
    assert_allclose(
        losses.GeneralizedDice(axis=(1, 2, 3))(x, y), losses.generalized_dice(x, y)
    )

    x = np.ones((4, 32, 32, 32, 1), dtype=np.float64)
    y = x.copy()
    x[:2, :10, 10:] = 0
    y[:2, :3, 20:] = 0
    y[3:, 10:] = 0
    # Dice is similar to generalized Dice for one class. The weight factor
    # makes the generalized form slightly different from Dice.
    gd = losses.generalized_dice(x, y, axis=(1, 2, 3)).numpy()
    dd = losses.dice(x, y, axis=(1, 2, 3, 4)).numpy()
    assert_allclose(gd, dd, rtol=1e-02)  # is this close enough?


def test_jaccard():
    x = np.zeros(4)
    y = np.zeros(4)
    out = losses.jaccard(x, y, axis=None).numpy()
    assert_allclose(out, 0)

    x = np.ones(4)
    y = np.ones(4)
    out = losses.jaccard(x, y, axis=None).numpy()
    assert_allclose(out, 0)

    x = [0.0, 0.0, 1.0, 1.0]
    y = [1.0, 1.0, 1.0, 1.0]
    out = losses.jaccard(x, y, axis=None).numpy()
    ref = scipy.spatial.distance.jaccard(x, y)
    assert_allclose(out, ref)

    x = [0.0, 0.0, 1.0, 1.0]
    y = [1.0, 1.0, 0.0, 0.0]
    out = losses.jaccard(x, y, axis=None).numpy()
    ref = scipy.spatial.distance.jaccard(x, y)
    assert_allclose(out, ref)
    assert_allclose(out, 1)

    x = np.ones((4, 32, 32, 32, 1), dtype=np.float32)
    y = x.copy()
    x[:2, :10, 10:] = 0
    y[:2, :3, 20:] = 0
    y[3:, 10:] = 0
    jaccards = np.empty(x.shape[0])
    for i in range(x.shape[0]):
        jaccards[i] = scipy.spatial.distance.jaccard(x[i].flatten(), y[i].flatten())
    assert_allclose(losses.jaccard(x, y, axis=(1, 2, 3, 4)), jaccards)
    assert_allclose(losses.Jaccard(axis=(1, 2, 3, 4))(x, y), jaccards.mean())
    assert_allclose(losses.Jaccard(axis=(1, 2, 3, 4))(y, x), jaccards.mean())


@pytest.mark.xfail
def test_tversky():
    # TODO: write the test
    assert False


@pytest.mark.xfail
def test_elbo():
    # TODO: write the test
    assert False


def test_wasserstein():
    x = np.zeros(4)
    y = np.zeros(4)
    out = losses.wasserstein(x, y)
    assert_allclose(out, 0)

    x = np.ones(4)
    y = np.ones(4)
    out = losses.wasserstein(x, y)
    assert_allclose(out, 1)

    x = np.array([0.0, -1.0, 1.0, -1.0])
    y = np.array([1.0, -1.0, 1.0, 1.0])
    out = losses.wasserstein(x, y)
    ref = [0.0, 1.0, 1.0, -1.0]
    assert_allclose(out, ref)

    x = np.array([0.0, 0.0, 1.0, 1.0])
    y = np.array([1.0, 1.0, 0.0, 0.0])
    out = losses.wasserstein(x, y)
    assert_allclose(out, 0)


def test_gradient_penalty():
    x = np.zeros(4)
    y = np.zeros(4)
    out = losses.gradient_penalty(x, y)
    assert_allclose(out, 10)

    x = np.ones(4)
    y = np.ones(4)
    out = losses.gradient_penalty(x, y)
    assert_allclose(out, 0.001)

    x = np.array([0.0, -1.0, 1.0, -1.0])
    y = np.array([1.0, -1.0, 1.0, 1.0])
    out = losses.gradient_penalty(x, y)
    ref = [1.0001e01, 1.0000e-03, 1.0000e-03, 1.0000e-03]
    assert_allclose(out, ref)

    x = np.array([0.0, 0.0, 1.0, 1.0])
    y = np.array([1.0, 1.0, 0.0, 0.0])
    out = losses.gradient_penalty(x, y)
    ref = [10.001, 10.001, 0.0, 0.0]
    assert_allclose(out, ref)


def test_get():
    if LooseVersion(tf.__version__) < LooseVersion("1.14.1-dev20190408"):
        assert losses.get("dice") is losses.dice
        assert losses.get("Dice") is losses.Dice
        assert losses.get("jaccard") is losses.jaccard
        assert losses.get("Jaccard") is losses.Jaccard
        assert losses.get("tversky") is losses.tversky
        assert losses.get("Tversky") is losses.Tversky
        assert losses.get("binary_crossentropy")
    else:
        assert losses.get("dice") is losses.dice
        assert isinstance(losses.get("Dice"), losses.Dice)
        assert losses.get("jaccard") is losses.jaccard
        assert isinstance(losses.get("Jaccard"), losses.Jaccard)
        assert losses.get("tversky") is losses.tversky
        assert isinstance(losses.get("Tversky"), losses.Tversky)
        assert losses.get("binary_crossentropy")
        assert losses.get("gradient_penalty") is losses.gradient_penalty
        assert losses.get("wasserstein") is losses.wasserstein
        assert isinstance(losses.get("Wasserstein"), losses.Wasserstein)



"""Attention U-net with inception layers.
Adapted from https://github.com/robinvvinod/unet
"""

from tensorflow.keras import layers
import tensorflow.keras.backend as K
from tensorflow.keras.models import Model

K.set_image_data_format("channels_last")


def expend_as(tensor, rep):
    # Anonymous lambda function to expand the specified axis by a factor of argument, rep.
    # If tensor has shape (512,512,N), lambda will return a tensor of shape
    # (512,512,N*rep), if specified axis=2

    my_repeat = layers.Lambda(
        lambda x, repnum: K.repeat_elements(x, repnum, axis=4),
        arguments={"repnum": rep},
    )(tensor)
    return my_repeat


def conv3d_block(
    input_tensor,
    n_filters,
    kernel_size=3,
    batchnorm=True,
    strides=1,
    dilation_rate=1,
    recurrent=1,
):
    # A wrapper of the Keras Conv3D block to serve as a building block for
    # downsampling layers
    # Includes options to use batch normalization, dilation and recurrence

    conv = layers.Conv3D(
        filters=n_filters,
        kernel_size=kernel_size,
        strides=strides,
        kernel_initializer="he_normal",
        padding="same",
        dilation_rate=dilation_rate,
    )(input_tensor)
    if batchnorm:
        conv = layers.BatchNormalization()(conv)
    output = layers.LeakyReLU(alpha=0.1)(conv)

    for _ in range(recurrent - 1):
        conv = layers.Conv3D(
            filters=n_filters,
            kernel_size=kernel_size,
            strides=1,
            kernel_initializer="he_normal",
            padding="same",
            dilation_rate=dilation_rate,
        )(output)
        if batchnorm:
            conv = layers.BatchNormalization()(conv)
        res = layers.LeakyReLU(alpha=0.1)(conv)
        output = layers.Add()([output, res])

    return output


def AttnGatingBlock(x, g, inter_shape):
    shape_x = K.int_shape(x)
    shape_g = K.int_shape(g)

    # Getting the gating signal to the same number of filters as the inter_shape
    phi_g = layers.Conv3D(
        filters=inter_shape, kernel_size=1, strides=1, padding="same"
    )(g)

    # Getting the x signal to the same shape as the gating signal
    theta_x = layers.Conv3D(
        filters=inter_shape,
        kernel_size=3,
        strides=(
            shape_x[1] // shape_g[1],
            shape_x[2] // shape_g[2],
            shape_x[3] // shape_g[3],
        ),
        padding="same",
    )(x)

    # Element-wise addition of the gating and x signals
    add_xg = layers.add([phi_g, theta_x])
    add_xg = layers.Activation("relu")(add_xg)

    # 1x1x1 convolution
    psi = layers.Conv3D(filters=1, kernel_size=1, padding="same")(add_xg)
    psi = layers.Activation("sigmoid")(psi)
    shape_sigmoid = K.int_shape(psi)

    # Upsampling psi back to the original dimensions of x signal
    upsample_sigmoid_xg = layers.UpSampling3D(
        size=(
            shape_x[1] // shape_sigmoid[1],
            shape_x[2] // shape_sigmoid[2],
            shape_x[3] // shape_sigmoid[3],
        )
    )(psi)

    # Expanding the filter axis to the number of filters in the original x signal
    upsample_sigmoid_xg = expend_as(upsample_sigmoid_xg, shape_x[4])

    # Element-wise multiplication of attention coefficients back onto original x signal
    attn_coefficients = layers.multiply([upsample_sigmoid_xg, x])

    # Final 1x1x1 convolution to consolidate attention signal to original x dimensions
    output = layers.Conv3D(
        filters=shape_x[4], kernel_size=1, strides=1, padding="same"
    )(attn_coefficients)
    output = layers.BatchNormalization()(output)
    return output


def transpose_block(
    input_tensor,
    skip_tensor,
    n_filters,
    kernel_size=3,
    strides=1,
    batchnorm=True,
    recurrent=1,
):
    # A wrapper of the Keras Conv3DTranspose block to serve as a building block
    # for upsampling layers

    shape_x = K.int_shape(input_tensor)
    shape_xskip = K.int_shape(skip_tensor)

    conv = layers.Conv3DTranspose(
        filters=n_filters,
        kernel_size=kernel_size,
        padding="same",
        strides=(
            shape_xskip[1] // shape_x[1],
            shape_xskip[2] // shape_x[2],
            shape_xskip[3] // shape_x[3],
        ),
        kernel_initializer="he_normal",
    )(input_tensor)
    conv = layers.LeakyReLU(alpha=0.1)(conv)

    act = conv3d_block(
        conv,
        n_filters=n_filters,
        kernel_size=kernel_size,
        strides=1,
        batchnorm=batchnorm,
        dilation_rate=1,
        recurrent=recurrent,
    )
    output = layers.Concatenate(axis=4)([act, skip_tensor])
    return output


# Use the functions provided in layers3D to build the network
def inception_block(
    input_tensor,
    n_filters,
    kernel_size=3,
    strides=1,
    batchnorm=True,
    recurrent=1,
    layers_list=[],
):
    # Inception-style convolutional block similar to InceptionNet
    # The first convolution follows the function arguments, while subsequent
    # inception convolutions follow the parameters in
    # argument, layers

    # layers is a nested list containing the different secondary inceptions in
    # the format of (kernel_size, dil_rate)

    # E.g => layers=[ [(3,1),(3,1)], [(5,1)], [(3,1),(3,2)] ]
    # This will implement 3 sets of secondary convolutions
    # Set 1 => 3x3 dil = 1 followed by another 3x3 dil = 1
    # Set 2 => 5x5 dil = 1
    # Set 3 => 3x3 dil = 1 followed by 3x3 dil = 2

    res = conv3d_block(
        input_tensor,
        n_filters=n_filters,
        kernel_size=kernel_size,
        strides=strides,
        batchnorm=batchnorm,
        dilation_rate=1,
        recurrent=recurrent,
    )

    temp = []
    for layer in layers_list:
        local_res = res
        for conv in layer:
            incep_kernel_size = conv[0]
            incep_dilation_rate = conv[1]
            local_res = conv3d_block(
                local_res,
                n_filters=n_filters,
                kernel_size=incep_kernel_size,
                strides=1,
                batchnorm=batchnorm,
                dilation_rate=incep_dilation_rate,
                recurrent=recurrent,
            )
        temp.append(local_res)

    temp = layers.concatenate(temp)
    res = conv3d_block(
        temp,
        n_filters=n_filters,
        kernel_size=1,
        strides=1,
        batchnorm=batchnorm,
        dilation_rate=1,
    )

    shortcut = conv3d_block(
        input_tensor,
        n_filters=n_filters,
        kernel_size=1,
        strides=strides,
        batchnorm=batchnorm,
        dilation_rate=1,
    )
    if batchnorm:
        shortcut = layers.BatchNormalization()(shortcut)

    output = layers.Add()([shortcut, res])
    return output


def attention_unet_with_inception(
    n_classes, input_shape, batch_size=None, n_filters=16, batchnorm=True
):
    # contracting path

    inputs = layers.Input(shape=input_shape, batch_size=batch_size)

    c0 = inception_block(
        inputs,
        n_filters=n_filters,
        batchnorm=batchnorm,
        strides=1,
        recurrent=2,
        layers_list=[[(3, 1), (3, 1)], [(3, 2)]],
    )  # 512x512x512

    c1 = inception_block(
        c0,
        n_filters=n_filters * 2,
        batchnorm=batchnorm,
        strides=2,
        recurrent=2,
        layers_list=[[(3, 1), (3, 1)], [(3, 2)]],
    )  # 256x256x256

    c2 = inception_block(
        c1,
        n_filters=n_filters * 4,
        batchnorm=batchnorm,
        strides=2,
        recurrent=2,
        layers_list=[[(3, 1), (3, 1)], [(3, 2)]],
    )  # 128x128x128

    c3 = inception_block(
        c2,
        n_filters=n_filters * 8,
        batchnorm=batchnorm,
        strides=2,
        recurrent=2,
        layers_list=[[(3, 1), (3, 1)], [(3, 2)]],
    )  # 64x64x64

    # bridge

    b0 = inception_block(
        c3,
        n_filters=n_filters * 16,
        batchnorm=batchnorm,
        strides=2,
        recurrent=2,
        layers_list=[[(3, 1), (3, 1)], [(3, 2)]],
    )  # 32x32x32

    # expansive path

    attn0 = AttnGatingBlock(c3, b0, n_filters * 16)
    u0 = transpose_block(
        b0, attn0, n_filters=n_filters * 8, batchnorm=batchnorm, recurrent=2
    )  # 64x64x64

    attn1 = AttnGatingBlock(c2, u0, n_filters * 8)
    u1 = transpose_block(
        u0, attn1, n_filters=n_filters * 4, batchnorm=batchnorm, recurrent=2
    )  # 128x128x128

    attn2 = AttnGatingBlock(c1, u1, n_filters * 4)
    u2 = transpose_block(
        u1, attn2, n_filters=n_filters * 2, batchnorm=batchnorm, recurrent=2
    )  # 256x256x256

    u3 = transpose_block(
        u2, c0, n_filters=n_filters, batchnorm=batchnorm, recurrent=2
    )  # 512x512x512

    outputs = layers.Conv3D(filters=1, kernel_size=1, strides=1)(u3)

    final_activation = "sigmoid" if n_classes == 1 else "softmax"
    outputs = layers.Activation(final_activation)(outputs)

    model = Model(inputs=[inputs], outputs=[outputs])
    return model


if __name__ == "__main__":
    model = attention_unet_with_inception(n_classes=1, input_shape=(256, 256, 256, 1))
    model.summary()



"""UNETR implementation in Tensorflow 2.0.

Adapted from https://www.kaggle.com/code/usharengaraju/tensorflow-unetr-w-b
"""
import math

import tensorflow as tf


class SingleDeconv3DBlock(tf.keras.layers.Layer):
    def __init__(self, filters):
        super(SingleDeconv3DBlock, self).__init__()
        self.block = tf.keras.layers.Conv3DTranspose(
            filters=filters,
            kernel_size=2,
            strides=2,
            padding="valid",
            output_padding=None,
        )

    def call(self, inputs):
        return self.block(inputs)


class SingleConv3DBlock(tf.keras.layers.Layer):
    def __init__(self, filters, kernel_size):
        super(SingleConv3DBlock, self).__init__()
        self.kernel = kernel_size
        self.res = tuple(map(lambda i: (i - 1) // 2, self.kernel))
        self.block = tf.keras.layers.Conv3D(
            filters=filters, kernel_size=kernel_size, strides=1, padding="same"
        )

    def call(self, inputs):
        return self.block(inputs)


class Conv3DBlock(tf.keras.layers.Layer):
    def __init__(self, filters, kernel_size=(3, 3, 3)):
        super(Conv3DBlock, self).__init__()
        self.a = tf.keras.Sequential(
            [
                SingleConv3DBlock(filters, kernel_size=kernel_size),
                tf.keras.layers.BatchNormalization(),
                tf.keras.layers.Activation("relu"),
            ]
        )

    def call(self, inputs):
        return self.a(inputs)


class Deconv3DBlock(tf.keras.layers.Layer):
    def __init__(self, filters, kernel_size=(3, 3, 3)):
        super(Deconv3DBlock, self).__init__()
        self.a = tf.keras.Sequential(
            [
                SingleDeconv3DBlock(filters=filters),
                SingleConv3DBlock(filters=filters, kernel_size=kernel_size),
                tf.keras.layers.BatchNormalization(),
                tf.keras.layers.Activation("relu"),
            ]
        )

    def call(self, inputs):
        return self.a(inputs)


class SelfAttention(tf.keras.layers.Layer):
    def __init__(self, num_heads, embed_dim, dropout):
        super(SelfAttention, self).__init__()

        self.num_attention_heads = num_heads
        self.attention_head_size = int(embed_dim / num_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = tf.keras.layers.Dense(self.all_head_size)
        self.key = tf.keras.layers.Dense(self.all_head_size)
        self.value = tf.keras.layers.Dense(self.all_head_size)

        self.out = tf.keras.layers.Dense(embed_dim)
        self.attn_dropout = tf.keras.layers.Dropout(dropout)
        self.proj_dropout = tf.keras.layers.Dropout(dropout)

        self.softmax = tf.keras.layers.Softmax()

        self.vis = False

    def transpose_for_scores(self, x):
        new_x_shape = list(
            x.shape[:-1] + (self.num_attention_heads, self.attention_head_size)
        )
        new_x_shape[0] = -1
        y = tf.reshape(x, new_x_shape)
        return tf.transpose(y, perm=[0, 2, 1, 3])

    def call(self, hidden_states):
        mixed_query_layer = self.query(hidden_states)
        mixed_key_layer = self.key(hidden_states)
        mixed_value_layer = self.value(hidden_states)

        query_layer = self.transpose_for_scores(mixed_query_layer)
        key_layer = self.transpose_for_scores(mixed_key_layer)
        value_layer = self.transpose_for_scores(mixed_value_layer)
        attention_scores = query_layer @ tf.transpose(key_layer, perm=[0, 1, 3, 2])
        attention_scores = attention_scores / math.sqrt(self.attention_head_size)
        attention_probs = self.softmax(attention_scores)
        weights = attention_probs if self.vis else None
        attention_probs = self.attn_dropout(attention_probs)

        context_layer = attention_probs @ value_layer
        context_layer = tf.transpose(context_layer, perm=[0, 2, 1, 3])
        new_context_layer_shape = list(context_layer.shape[:-2] + (self.all_head_size,))
        new_context_layer_shape[0] = -1
        context_layer = tf.reshape(context_layer, new_context_layer_shape)
        attention_output = self.out(context_layer)
        attention_output = self.proj_dropout(attention_output)
        return attention_output, weights


class Mlp(tf.keras.layers.Layer):
    def __init__(self, output_features, drop=0.0):
        super(Mlp, self).__init__()
        self.a = tf.keras.layers.Dense(units=output_features, activation=tf.nn.gelu)
        self.b = tf.keras.layers.Dropout(drop)

    def call(self, inputs):
        x = self.a(inputs)
        return self.b(x)


class PositionwiseFeedForward(tf.keras.layers.Layer):
    def __init__(self, d_model=768, d_ff=2048, dropout=0.1):
        super(PositionwiseFeedForward, self).__init__()
        self.a = tf.keras.layers.Dense(units=d_ff)
        self.b = tf.keras.layers.Dense(units=d_model)
        self.c = tf.keras.layers.Dropout(dropout)

    def call(self, inputs):
        return self.b(self.c(tf.nn.relu(self.a(inputs))))


# embeddings, projection_dim=embed_dim
class PatchEmbedding(tf.keras.layers.Layer):
    def __init__(self, cube_size, patch_size, embed_dim):
        super(PatchEmbedding, self).__init__()
        self.num_of_patches = int(
            (cube_size[0] * cube_size[1] * cube_size[2])
            / (patch_size * patch_size * patch_size)
        )
        self.patch_size = patch_size
        self.size = patch_size
        self.embed_dim = embed_dim

        self.projection = tf.keras.layers.Dense(embed_dim)

        self.clsToken = tf.Variable(
            tf.keras.initializers.GlorotNormal()(shape=(1, 512, embed_dim)),
            trainable=True,
        )

        self.positionalEmbedding = tf.keras.layers.Embedding(
            self.num_of_patches, embed_dim
        )
        self.patches = None
        self.lyer = tf.keras.layers.Conv3D(
            filters=self.embed_dim,
            kernel_size=self.patch_size,
            strides=self.patch_size,
            padding="valid",
        )
        # embedding - basically is adding numerical embedding to the layer along with an extra dim

    def call(self, inputs):
        patches = self.lyer(inputs)
        patches = tf.reshape(
            patches, (tf.shape(inputs)[0], -1, self.size * self.size * 3)
        )
        patches = self.projection(patches)
        positions = tf.range(0, self.num_of_patches, 1)[tf.newaxis, ...]
        positionalEmbedding = self.positionalEmbedding(positions)
        patches = patches + positionalEmbedding

        return patches, positionalEmbedding


# transformerblock
class TransformerLayer(tf.keras.layers.Layer):
    def __init__(self, embed_dim, num_heads, dropout, cube_size, patch_size):
        super(TransformerLayer, self).__init__()

        self.attention_norm = tf.keras.layers.LayerNormalization(epsilon=1e-6)

        self.mlp_norm = tf.keras.layers.LayerNormalization(epsilon=1e-6)

        # embed_dim/no-of_heads
        self.mlp_dim = int(
            (cube_size[0] * cube_size[1] * cube_size[2])
            / (patch_size * patch_size * patch_size)
        )

        self.mlp = PositionwiseFeedForward(embed_dim, 2048)
        self.attn = SelfAttention(num_heads, embed_dim, dropout)

    def call(self, x, training=True):
        h = x
        x = self.attention_norm(x)
        x, weights = self.attn(x)
        x = x + h
        h = x

        x = self.mlp_norm(x)
        x = self.mlp(x)

        x = x + h

        return x, weights


class TransformerEncoder(tf.keras.layers.Layer):
    def __init__(
        self,
        embed_dim,
        num_heads,
        cube_size,
        patch_size,
        num_layers=12,
        dropout=0.1,
        extract_layers=[3, 6, 9, 12],
    ):
        super(TransformerEncoder, self).__init__()
        #  embed_dim, num_heads ,dropout, cube_size, patch_size
        self.embeddings = PatchEmbedding(cube_size, patch_size, embed_dim)
        self.extract_layers = extract_layers
        self.encoders = [
            TransformerLayer(embed_dim, num_heads, dropout, cube_size, patch_size)
            for _ in range(num_layers)
        ]

    def call(self, inputs, training=True):
        extract_layers = []
        x = inputs
        x, _ = self.embeddings(x)

        for depth, layer in enumerate(self.encoders):
            x, _ = layer(x, training=training)
            if depth + 1 in self.extract_layers:
                extract_layers.append(x)

        return extract_layers


class UNETR(tf.keras.Model):
    def __init__(
        self,
        img_shape=(96, 96, 96),
        input_dim=3,
        output_dim=3,
        embed_dim=768,
        patch_size=16,
        num_heads=12,
        dropout=0.1,
    ):
        super(UNETR, self).__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.embed_dim = embed_dim
        self.img_shape = img_shape
        self.patch_size = patch_size
        self.num_heads = num_heads
        self.dropout = dropout
        self.num_layers = 12
        self.ext_layers = [3, 6, 9, 12]

        self.patch_dim = [int(x / patch_size) for x in img_shape]
        self.transformer = TransformerEncoder(
            self.embed_dim,
            self.num_heads,
            self.img_shape,
            self.patch_size,
            self.num_layers,
            self.dropout,
            self.ext_layers,
        )

        # U-Net Decoder
        self.decoder0 = tf.keras.Sequential(
            [Conv3DBlock(32, (3, 3, 3)), Conv3DBlock(64, (3, 3, 3))]
        )

        self.decoder3 = tf.keras.Sequential(
            [Deconv3DBlock(512), Deconv3DBlock(256), Deconv3DBlock(128)]
        )

        self.decoder6 = tf.keras.Sequential([Deconv3DBlock(512), Deconv3DBlock(256)])

        self.decoder9 = Deconv3DBlock(512)

        self.decoder12_upsampler = SingleDeconv3DBlock(512)

        self.decoder9_upsampler = tf.keras.Sequential(
            [
                Conv3DBlock(512),
                Conv3DBlock(512),
                Conv3DBlock(512),
                SingleDeconv3DBlock(256),
            ]
        )

        self.decoder6_upsampler = tf.keras.Sequential(
            [Conv3DBlock(256), Conv3DBlock(256), SingleDeconv3DBlock(128)]
        )

        self.decoder3_upsampler = tf.keras.Sequential(
            [Conv3DBlock(128), Conv3DBlock(128), SingleDeconv3DBlock(64)]
        )

        self.decoder0_header = tf.keras.Sequential(
            [Conv3DBlock(64), Conv3DBlock(64), SingleConv3DBlock(output_dim, (1, 1, 1))]
        )

    def call(self, x):
        z = self.transformer(x)
        z0, z3, z6, z9, z12 = x, z[0], z[1], z[2], z[3]
        z3 = tf.reshape(
            tf.transpose(z3, perm=[0, 2, 1]), [-1, *self.patch_dim, self.embed_dim]
        )
        z6 = tf.reshape(
            tf.transpose(z6, perm=[0, 2, 1]), [-1, *self.patch_dim, self.embed_dim]
        )
        z9 = tf.reshape(
            tf.transpose(z9, perm=[0, 2, 1]), [-1, *self.patch_dim, self.embed_dim]
        )
        z12 = tf.reshape(
            tf.transpose(z12, perm=[0, 2, 1]), [-1, *self.patch_dim, self.embed_dim]
        )
        z12 = self.decoder12_upsampler(z12)
        z9 = self.decoder9(z9)
        z9 = self.decoder9_upsampler(tf.concat([z9, z12], 4))
        z6 = self.decoder6(z6)
        z6 = self.decoder6_upsampler(tf.concat([z6, z9], 4))
        z3 = self.decoder3(z3)
        z3 = self.decoder3_upsampler(tf.concat([z3, z6], 4))
        z0 = self.decoder0(z0)
        output = self.decoder0_header(tf.concat([z0, z3], 4))
        return output

    # def model(self):
    #     x = tf.keras.layers.Input(shape=(96, 96, 96, 3))
    #     return tf.keras.Model(inputs=[x], outputs=self.call(x))


def unetr(
    n_classes=1,
    input_shape=(96, 96, 96, 3),
    embed_dim=768,
    patch_size=16,
    num_heads=12,
    dropout=0.1,
):
    *img_shape, input_dim = input_shape

    input = tf.keras.layers.Input([*img_shape, input_dim], name="input_image")

    z = UNETR(
        img_shape=img_shape,
        input_dim=input_dim,
        output_dim=n_classes,
        embed_dim=embed_dim,
        patch_size=patch_size,
        num_heads=num_heads,
        dropout=dropout,
    )(input)

    final_activation = "sigmoid" if n_classes == 1 else "softmax"
    output = tf.keras.layers.Activation(final_activation)(z)

    return tf.keras.Model(inputs=[input], outputs=[output])


if __name__ == "__main__":
    input_shape = (96, 96, 96, 3)
    sub1 = unetr(input_shape=input_shape, n_classes=1)
    sub1.summary()



from pprint import pprint

from .attention_unet import attention_unet
from .attention_unet_with_inception import attention_unet_with_inception
from .autoencoder import autoencoder
from .bayesian_meshnet import variational_meshnet
from .dcgan import dcgan
from .highresnet import highresnet
from .meshnet import meshnet
from .progressiveae import progressiveae
from .progressivegan import progressivegan
from .unet import unet
from .unetr import unetr

__all__ = ["get", "list_available_models"]

_models = {
    "highresnet": highresnet,
    "meshnet": meshnet,
    "unet": unet,
    "autoencoder": autoencoder,
    "progressivegan": progressivegan,
    "progressiveae": progressiveae,
    "dcgan": dcgan,
    "attention_unet": attention_unet,
    "attention_unet_with_inception": attention_unet_with_inception,
    "unetr": unetr,
    "variational_meshnet": variational_meshnet,
}


def get(name):
    """Return callable that creates a particular `tf.keras.Model`.

    Parameters
    ----------
    name: str, the name of the model (case-insensitive).

    Returns
    -------
    Callable, which instantiates a `tf.keras.Model` object.
    """
    if not isinstance(name, str):
        raise ValueError("Model name must be a string.")

    try:
        return _models[name.lower()]
    except KeyError:
        avail = ", ".join(_models.keys())
        raise ValueError(
            "Unknown model: '{}'. Available models are {}.".format(name, avail)
        )


def available_models():
    return list(_models)


def list_available_models():
    pprint(available_models())



import functools
import math
import os
from pathlib import Path

from joblib import Parallel, delayed
import numpy as np
import skimage.transform
import tensorflow as tf

from .io import read_volume
from .utils import get_num_parallel

_TFRECORDS_DTYPE = "float32"


def __writer_func(iterator_filename, map_fn):
    iterator, filename = iterator_filename
    map_fn(protobuf_iterator=iterator, filename=filename)


def write(
    features_labels,
    filename_template,
    examples_per_shard,
    to_ras=True,
    compressed=True,
    processes=None,
    chunksize=1,
    multi_resolution=False,
    resolutions=None,
    verbose=1,
):
    """Write features and labels to TFRecord files.

    This method supports the use of scalar or n-D labels. All labels, however, must be
    scalar, or all labels must be n-D. If labels are n-D, their shape must be the same
    as the shape of features.

    Parameters
    ----------
    features_labels: nested sequence, pairs of features and labels. If the labels are
        n-D, each pair should be `(path_to_feature, path_to_label)`. If labels are
        scalar, each pair should be `(path_to_feature, scalar_label)`.
    filename_template: str, the path to which TFRecord files should be written. A string
        formatting key `shard` should be included to indicate the unique TFRecord file
        when writing to multiple TFRecord files. For example,
        `data_shard-{shard:03d}.tfrec`.
    examples_per_shard: int, number of pairs of `(feature, label)` per TFRecord file.
    to_ras: boolean, reorient volumes to RAS before writing to TFRecord file.
    compressed: boolean, if `True`, gzip compress TFRecord file. Highly recommended.
    processes: int, number of processes for multiprocessing. This is useful when
        writing to multiple TFRecord files (i.e.,
        `examples_per_shard` < `len(features_labels)`). If `None`, uses all available
        cores.
    chunksize: int, multiprocessing chunksize.
    multi_resolution: boolean, if `True`, different tfrecords for each resolution in each shard
    resolutions: list of ints, if multi_resolution is `True`, set resolutions for
        which tfrecords are created. For example, [4, 8, 16, 32, 64, 128, 256]
    verbose: int, if 1, print progress bar. If 0, print nothing.
    """
    n_examples = len(features_labels)
    n_shards = math.ceil(n_examples / examples_per_shard)
    shards = np.array_split(features_labels, n_shards)

    # Test that the `filename_template` has a `shard` formatting key.
    try:
        filename_template.format(shard=0)
    except Exception:
        raise ValueError(
            "`filename_template` must include a string formatting key 'shard' that"
            " accepts an integer."
        )
    # Test if resolutions is set
    if multi_resolution and not isinstance(resolutions, list):
        raise ValueError(
            "`resolutions` must be set with a list of ints indicating resolutions for tfrecords"
        )

    # This is the object that returns a protocol buffer string of the feature and label
    # on each iteration. It is pickle-able, unlike a generator.
    proto_iterators = [
        _ProtoIterator(s, multi_resolution=multi_resolution, resolutions=resolutions)
        for s in shards
    ]
    # Set up positional arguments for the core writer function.
    iterable = [
        (p, filename_template.format(shard=j)) for j, p in enumerate(proto_iterators)
    ]
    # Set keyword arguments so the resulting function accepts one positional argument.
    map_fn = functools.partial(
        _write_tfrecords,
        compressed=True,
        multi_resolution=multi_resolution,
        resolutions=resolutions,
    )

    if processes is None:
        processes = get_num_parallel()
    Parallel(n_jobs=processes, verbose=10)(
        delayed(__writer_func)(val, map_fn) for val in iterable
    )
    from joblib.externals.loky import get_reusable_executor

    get_reusable_executor().shutdown(wait=True)


def parse_example_fn(volume_shape, scalar_labels=False):
    """Return function that can be used to read TFRecord file into tensors.

    Parameters
    ----------
    volume_shape: sequence, the shape of the feature data. If `scalar_labels` is `False`,
        this also corresponds to the shape of the label data.
    scalar_labels: boolean, if `True`, label is a scalar. If `False`, label must be the
        same shape as feature data.

    Returns
    -------
    Function with which a TFRecord file can be parsed.
    """

    @tf.function
    def parse_example(serialized):
        """Parse one example from a TFRecord file made with Nobrainer.

        Parameters
        ----------
        serialized: str, serialized proto message.

        Returns
        -------
        Tuple of two tensors. If `scalar_labels` is `False`, both tensors have shape
        `volume_shape`. Otherwise, the first tensor has shape `volume_shape`, and the
        second is a scalar tensor.
        """
        features = {
            "feature/shape": tf.io.FixedLenFeature(shape=[], dtype=tf.string),
            "feature/value": tf.io.FixedLenFeature(shape=[], dtype=tf.string),
            "label/value": tf.io.FixedLenFeature(shape=[], dtype=tf.string),
            "label/rank": tf.io.FixedLenFeature(shape=[], dtype=tf.int64),
        }
        e = tf.io.parse_single_example(serialized=serialized, features=features)
        x = tf.io.decode_raw(e["feature/value"], _TFRECORDS_DTYPE)
        y = tf.io.decode_raw(e["label/value"], _TFRECORDS_DTYPE)
        # TODO: this line does not work. The shape cannot be determined
        # dynamically... for now.
        # xshape = tf.cast(
        #     tf.io.decode_raw(e["feature/shape"], _TFRECORDS_DTYPE), tf.int32)
        x = tf.reshape(x, shape=volume_shape)
        if not scalar_labels:
            y = tf.reshape(y, shape=volume_shape)
        else:
            y = tf.reshape(y, shape=[1])
        return x, y

    return parse_example


def _bytes_feature(value):
    """Returns a bytes_list from a string / byte."""
    if isinstance(value, type(tf.constant(0))):
        value = value.numpy()  # BytesList won't unpack a string from an EagerTensor.
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def _float_feature(value):
    """Returns a float_list from a float / double."""
    return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))


def _int64_feature(value):
    """Returns an int64_list from a bool / enum / int / uint."""
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


def _dtype_to_bytes(dtype):
    if isinstance(dtype, str):
        return dtype.encode("utf-8")
    try:
        s = str(dtype.name)
    except AttributeError:
        s = str(dtype)
    return s.encode("utf-8")


def _get_feature_dict(x, affine):
    x = np.asarray(x)
    x = x.astype(_TFRECORDS_DTYPE)

    feature = {
        "feature/value": _bytes_feature(x.tobytes()),
        "feature/dtype": _bytes_feature(_dtype_to_bytes(x.dtype)),
        "feature/rank": _int64_feature(x.ndim),
    }

    for j, s in enumerate(x.shape):
        feature["feature/shape/dim{}".format(j)] = _int64_feature(s)

    if x.ndim:
        feature["feature/shape"] = _bytes_feature(
            np.array(x.shape).astype(_TFRECORDS_DTYPE).tobytes()
        )

    affine = np.asarray(affine)
    affine = affine.astype(_TFRECORDS_DTYPE)
    feature.update(
        {
            "feature/affine/value": _bytes_feature(affine.tobytes()),
            "feature/affine/dtype": _bytes_feature(_dtype_to_bytes(affine.dtype)),
            "feature/affine/rank": _int64_feature(affine.ndim),
        }
    )
    for j, s in enumerate(affine.shape):
        feature["feature/affine/shape/dim{}".format(j)] = _int64_feature(s)

    return feature


def _get_label_dict(y, affine=None):
    y = np.asarray(y)
    y = y.astype(_TFRECORDS_DTYPE)

    label = {
        "label/value": _bytes_feature(y.tobytes()),
        "label/dtype": _bytes_feature(_dtype_to_bytes(y.dtype)),
        "label/rank": _int64_feature(y.ndim),
    }

    for j, s in enumerate(y.shape):
        label["label/shape/dim{}".format(j)] = _int64_feature(s)
    if y.ndim:
        label["feature/shape"] = _bytes_feature(
            np.array(y.shape).astype(_TFRECORDS_DTYPE).tobytes()
        )

    if affine is None and y.ndim != 0:
        raise ValueError("Affine is required when label is not scalar.")

    if affine is not None:
        affine = np.asarray(affine)
        affine = affine.astype(_TFRECORDS_DTYPE)
        label.update(
            {
                "label/affine/value": _bytes_feature(affine.tobytes()),
                "label/affine/dtype": _bytes_feature(_dtype_to_bytes(affine.dtype)),
                "label/affine/rank": _int64_feature(affine.ndim),
            }
        )
        for j, s in enumerate(affine.shape):
            label["label/affine/shape/dim{}".format(j)] = _int64_feature(s)

    return label


def _to_proto(feature, label, feature_affine, label_affine=None):
    """Return instance of `tf.train.Example` of feature and label."""
    d = _get_feature_dict(x=feature, affine=feature_affine)
    d.update(_get_label_dict(y=label, affine=label_affine))
    return tf.train.Example(features=tf.train.Features(feature=d))


class _ProtoIterator:
    """Iterator of protobuf strings.

    Parameters
    ----------

    This custom iterator is used instead of a generator for use in
    multiprocessing. Generators cannot be pickled, and so cannot be used in
    multiprocessing workflows. Please see
    https://stackoverflow.com/a/7180424/5666087 for more information.
    """

    def __init__(
        self, features_labels, to_ras=True, multi_resolution=False, resolutions=None
    ):
        self.features_labels = features_labels
        self.to_ras = to_ras
        self.multi_resolution = multi_resolution
        self.resolutions = resolutions

        # Try to "intelligently" deduce if the labels are scalars or not.
        # An alternative here would be to check if these point to existing
        # files, though it is possible to have existing filenames that
        # are integers or floats.
        labels = [y for _, y in features_labels]
        self.scalar_labels = _labels_all_scalar(labels)
        self._j = 0

        no_exist = []
        for x, _ in self.features_labels:
            if not Path(x).exists():
                no_exist.append(x)
        if no_exist:
            raise ValueError("Some files do not exist: {}".format(", ".join(no_exist)))

        if not self.scalar_labels:
            no_exist = []
            for _, y in self.features_labels:
                if not Path(y).exists():
                    no_exist.append(y)
            if no_exist:
                raise ValueError(
                    "Some files do not exist: {}".format(", ".join(no_exist))
                )

    def __iter__(self):
        self._j = 0
        return self

    def __next__(self):
        index = self._j
        serialized = self._serialize(index)
        self._j += 1
        return serialized

    def _serialize(self, index):
        try:
            x, y = self.features_labels[index]
        except IndexError:
            raise StopIteration
        x, affine_x = read_volume(
            x, return_affine=True, dtype=_TFRECORDS_DTYPE, to_ras=self.to_ras
        )
        if self.multi_resolution:
            # only scalar label
            if not self.scalar_labels:
                y = 0
            proto_dict = {}
            for resolution in self.resolutions[::-1]:
                x_res = skimage.transform.resize(
                    x,
                    output_shape=(resolution, resolution, resolution),
                    order=1,  # linear
                    mode="constant",
                    preserve_range=True,
                    anti_aliasing=True,
                )
                # TODO: Affine should be adjusted when resizing
                proto = _to_proto(
                    feature=x_res, label=y, feature_affine=affine_x, label_affine=None
                )
                proto_dict[resolution] = proto.SerializeToString()

            return proto_dict
        else:
            label_affine = None
            if not self.scalar_labels:
                y, label_affine = read_volume(
                    y, return_affine=True, dtype=_TFRECORDS_DTYPE, to_ras=self.to_ras
                )
            proto = _to_proto(
                feature=x,
                label=y,
                feature_affine=affine_x,
                label_affine=label_affine,
            )
            return proto.SerializeToString()


def _write_tfrecords(
    protobuf_iterator,
    filename,
    compressed=True,
    multi_resolution=False,
    resolutions=None,
):
    """
    protobuf_iterator: iterator, iterator which yields protocol-buffer serialized
        strings.
    """
    if compressed:
        options = tf.io.TFRecordOptions(compression_type="GZIP")
    else:
        options = None
    if not multi_resolution:
        with tf.io.TFRecordWriter(filename, options=options) as f:
            for proto_string in protobuf_iterator:
                f.write(proto_string)
    else:
        tf_record_writers = {}
        filenames = [
            "{0}-res-{2:03d}{1}".format(*os.path.splitext(filename) + (res,))
            for res in resolutions
        ]
        for resolution, filename in zip(resolutions, filenames):
            tf_record_writers[resolution] = tf.io.TFRecordWriter(
                filename, options=options
            )
        for proto_string_dict in protobuf_iterator:
            for resolution in resolutions:
                tf_record_writers[resolution].write(proto_string_dict[resolution])
        [tf_record_writers[writer].close() for writer in tf_record_writers]


def _is_int_or_float(value):
    if isinstance(value, (int, float)):
        return True
    try:
        float(value)
        return True
    except (TypeError, ValueError):
        pass
    try:
        int(value)
        return True
    except (TypeError, ValueError):
        pass
    return False


def _labels_all_scalar(labels):
    scalars = list(map(_is_int_or_float, labels))
    if any(scalars) and not all(scalars):
        raise ValueError(
            "Some labels were detected as scalars, while others were not."
            " Labels must be all scalars or all filenames of volumes."
        )
    return all(scalars)



import numpy as np
import tensorflow as tf

from ..padding import ZeroPadding3DChannels


def test_zeropadding3dchannels():
    # This test function is a much shorter version of
    # `tensorflow.python.keras.testing_utils.layer_test`.
    input_data_shape = (4, 32, 32, 32, 1)
    input_data = 10 * np.random.random(input_data_shape)

    x = tf.keras.layers.Input(shape=input_data_shape[1:], dtype=input_data.dtype)
    y = ZeroPadding3DChannels(4)(x)
    model = tf.keras.Model(x, y)

    actual_output = model.predict(input_data)
    actual_output_shape = actual_output.shape
    assert actual_output_shape == (4, 32, 32, 32, 9)
    assert not actual_output[..., :4].any()
    assert actual_output[..., 4].any()
    assert not actual_output[..., 5:].any()

    return actual_output



import numpy as np
import tensorflow as tf

from nobrainer import intensity_transforms


def test_addGaussianNoise():
    shape = (10, 10, 10)
    x = np.ones(shape).astype(np.float32)
    y = np.random.randint(0, 2, size=shape).astype(np.float32)
    x_out = intensity_transforms.addGaussianNoise(x, noise_mean=0.0, noise_std=1)
    x_out = x_out.numpy()
    assert x_out.shape == x.shape
    assert np.sum(x_out - x) != 0

    # test if x and y undergoes same noiseshift
    x_out, y_out = intensity_transforms.addGaussianNoise(
        x, y, trans_xy=True, noise_mean=0.0, noise_std=1
    )
    x_out = x_out.numpy()
    y_out = y_out.numpy()
    noise_y = y_out - y
    noise_x = x_out - x
    assert x_out.shape == x.shape
    assert y_out.shape == y.shape
    assert np.sum(noise_x - noise_y) < 1e-5

    # test sending y, but not transforming it
    x_out, y_out = intensity_transforms.addGaussianNoise(
        x, y, trans_xy=False, noise_mean=0.0, noise_std=1
    )
    x_out = x_out.numpy()
    assert x_out.shape == x.shape
    assert np.sum(x_out - x) != 0
    assert y_out.shape == y.shape
    np.testing.assert_array_equal(y_out, y)


def test_minmaxIntensityScaling():
    x = np.random.rand(10, 10, 10).astype(np.float32)
    y = np.random.randint(0, 2, size=(10, 10, 10)).astype(np.float32)
    x_out, y_out = intensity_transforms.minmaxIntensityScaling(x, y, trans_xy=True)
    x_out = x_out.numpy()
    y_out = y_out.numpy()
    assert x_out.min() - 0.0 < 1e-5
    assert y_out.min() - 0.0 < 1e-5
    assert 1 - x_out.max() < 1e-5
    assert 1 - y_out.max() < 1e-5

    x_out, y_out = intensity_transforms.minmaxIntensityScaling(x, y, trans_xy=False)
    x_out = x_out.numpy()
    assert x_out.min() - 0.0 < 1e-5
    assert 1 - x_out.max() < 1e-5
    np.testing.assert_array_equal(y_out, y)


def test_customIntensityScaling():
    x = np.random.rand(10, 10, 10).astype(np.float32)
    y = np.random.randint(0, 2, size=(10, 10, 10)).astype(np.float32)
    x_out, y_out = intensity_transforms.customIntensityScaling(
        x, y, trans_xy=True, scale_x=[0, 100], scale_y=[0, 3]
    )
    x_out = x_out.numpy()
    y_out = y_out.numpy()
    assert x_out.min() - 0.0 < 1e-5
    assert y_out.min() - 0.0 < 1e-5
    assert 100 - x_out.max() < 1e-5
    assert 3 - y_out.max() < 1e-5

    x_out, y_out = intensity_transforms.customIntensityScaling(
        x, y, trans_xy=False, scale_x=[0, 100], scale_y=[0, 3]
    )
    x_out = x_out.numpy()
    assert x_out.min() - 0.0 < 1e-5
    assert 100 - x_out.max() < 1e-5
    np.testing.assert_array_equal(y_out, y)


def test_intensityMasking():
    mask_x = np.array([[[0, 0, 0], [0, 1, 0], [0, 0, 0]]])
    x = np.array([[[1, 1, 1], [2, 2, 2], [3, 3, 3]], [[4, 4, 4], [5, 5, 5], [6, 6, 6]]])
    expected = np.array(
        [[[0, 0, 0], [0, 2, 0], [0, 0, 0]], [[0, 0, 0], [0, 5, 0], [0, 0, 0]]]
    )
    results = intensity_transforms.intensityMasking(x, mask_x=mask_x)
    results = tf.squeeze(results)
    np.testing.assert_allclose(results.numpy(), expected)

    y = np.random.rand(*x.shape)
    results, y_out = intensity_transforms.intensityMasking(
        x, mask_x=mask_x, y=y, trans_xy=False
    )
    results = tf.squeeze(results)
    np.testing.assert_allclose(results.numpy(), expected)
    np.testing.assert_array_equal(y_out, y)


def test_contrastAdjust():
    gamma = 1.5
    epsilon = 1e-7
    x = np.array([[[1, 1, 1], [2, 2, 2], [3, 3, 3]], [[4, 4, 4], [5, 5, 5], [6, 6, 6]]])
    x_range = x.max() - x.min()
    expected = (
        np.power(((x - x.min()) / float(x_range + epsilon)), gamma) * x_range + x.min()
    )
    results = intensity_transforms.contrastAdjust(x, gamma=1.5)
    np.testing.assert_allclose(expected, results.numpy(), rtol=1e-05)

    y = np.random.rand(*x.shape)
    results, y_out = intensity_transforms.contrastAdjust(
        x, y, trans_xy=False, gamma=1.5
    )
    np.testing.assert_allclose(expected, results.numpy(), rtol=1e-05)
    np.testing.assert_array_equal(y_out, y)



"""Model definition for UNet.
"""

import tensorflow as tf
from tensorflow.keras import layers


def unet(
    n_classes,
    input_shape,
    activation="relu",
    batchnorm=False,
    batch_size=None,
    name="unet",
):
    """
    Instantiate a 3D UNet Architecture
    UNet model: a 3D deep neural network model from
    https://arxiv.org/abs/1606.06650
    Args:
    n_classes(int): number of classes
    input_shape(tuple):four ints representing the shape of 3D input
    activation(str): all tf.keras.activations are allowed
    batch_size(int): batch size.
    """

    conv_kwds = {
        "kernel_size": (3, 3, 3),
        "activation": None,
        "padding": "same",
        # 'kernel_regularizer': tf.keras.regularizers.l2(0.001),
    }

    conv_transpose_kwds = {
        "kernel_size": (2, 2, 2),
        "strides": 2,
        "padding": "same",
        # 'kernel_regularizer': tf.keras.regularizers.l2(0.001),
    }

    n_base_filters = 16
    inputs = layers.Input(shape=input_shape, batch_size=batch_size)

    # Begin analysis path (encoder).

    x = layers.Conv3D(n_base_filters, **conv_kwds)(inputs)
    if batchnorm:
        x = layers.BatchNormalization()(x)
    x = layers.Activation(activation)(x)
    x = layers.Conv3D(n_base_filters * 2, **conv_kwds)(x)
    if batchnorm:
        x = layers.BatchNormalization()(x)
    skip_1 = x = layers.Activation(activation)(x)
    x = layers.MaxPool3D(2)(x)

    x = layers.Conv3D(n_base_filters * 2, **conv_kwds)(x)
    if batchnorm:
        x = layers.BatchNormalization()(x)
    x = layers.Activation(activation)(x)
    x = layers.Conv3D(n_base_filters * 4, **conv_kwds)(x)
    if batchnorm:
        x = layers.BatchNormalization()(x)
    skip_2 = x = layers.Activation(activation)(x)
    x = layers.MaxPool3D(2)(x)

    x = layers.Conv3D(n_base_filters * 4, **conv_kwds)(x)
    if batchnorm:
        x = layers.BatchNormalization()(x)
    x = layers.Activation(activation)(x)
    x = layers.Conv3D(n_base_filters * 8, **conv_kwds)(x)
    if batchnorm:
        x = layers.BatchNormalization()(x)
    skip_3 = x = layers.Activation(activation)(x)
    x = layers.MaxPool3D(2)(x)

    x = layers.Conv3D(n_base_filters * 8, **conv_kwds)(x)
    if batchnorm:
        x = layers.BatchNormalization()(x)
    x = layers.Activation(activation)(x)
    x = layers.Conv3D(n_base_filters * 16, **conv_kwds)(x)
    if batchnorm:
        x = layers.BatchNormalization()(x)
    x = layers.Activation(activation)(x)

    # End analysis path (encoder).
    # Begin synthesis path (decoder).

    x = layers.Conv3DTranspose(n_base_filters * 16, **conv_transpose_kwds)(x)

    x = layers.Concatenate(axis=-1)([skip_3, x])
    x = layers.Conv3D(n_base_filters * 8, **conv_kwds)(x)
    if batchnorm:
        x = layers.BatchNormalization()(x)
    x = layers.Activation(activation)(x)
    x = layers.Conv3D(n_base_filters * 8, **conv_kwds)(x)
    if batchnorm:
        x = layers.BatchNormalization()(x)
    x = layers.Activation(activation)(x)

    x = layers.Conv3DTranspose(n_base_filters * 8, **conv_transpose_kwds)(x)

    x = layers.Concatenate(axis=-1)([skip_2, x])
    x = layers.Conv3D(n_base_filters * 4, **conv_kwds)(x)
    if batchnorm:
        x = layers.BatchNormalization()(x)
    x = layers.Activation(activation)(x)
    x = layers.Conv3D(n_base_filters * 4, **conv_kwds)(x)
    if batchnorm:
        x = layers.BatchNormalization()(x)
    x = layers.Activation(activation)(x)

    x = layers.Conv3DTranspose(n_base_filters * 4, **conv_transpose_kwds)(x)

    x = layers.Concatenate(axis=-1)([skip_1, x])
    x = layers.Conv3D(n_base_filters * 2, **conv_kwds)(x)
    if batchnorm:
        x = layers.BatchNormalization()(x)
    x = layers.Activation(activation)(x)
    x = layers.Conv3D(n_base_filters * 2, **conv_kwds)(x)
    if batchnorm:
        x = layers.BatchNormalization()(x)
    x = layers.Activation(activation)(x)

    x = layers.Conv3D(filters=n_classes, kernel_size=1)(x)

    final_activation = "sigmoid" if n_classes == 1 else "softmax"
    x = layers.Activation(final_activation)(x)

    return tf.keras.Model(inputs=inputs, outputs=x, name=name)



# Model definition of bayesian adaptation of the Vnet model
# from https://arxiv.org/pdf/1606.04797.pdf
from tensorflow.keras.layers import Input, MaxPooling3D, UpSampling3D, concatenate
from tensorflow.keras.models import Model
import tensorflow_probability as tfp

from ..bayesian_utils import normal_prior
from ..layers.groupnorm import GroupNormalization


def down_stage(
    inputs,
    filters,
    prior_fn,
    kernel_posterior_fn,
    kld,
    kernel_size=3,
    activation="relu",
    padding="SAME",
):
    """encoding blocks of the Bayesian VNet model.

    Parameters
    ----------
    inputs: tf.layer for encoding stage.
    filters: list or tuple of four ints, the shape of the input data. Omit
        the batch dimension, and include the number of channels.
    kld: a func to compute KL Divergence loss, default is set None.
    KLD can be set as (lambda q, p, ignore: kl_lib.kl_divergence(q, p))
    prior_fn: a func to initialize priors distributions
    kernel_posterior_fn:a func to initlaize kernel posteriors
        (loc, scale and weightnorms)
    kernal_size: int, size of the kernel of conv layers. Default kernel size
        is set to be 3.
    activation: str or optimizer object, the non-linearity to use. All
        tf.activations are allowed to use.

    Returns
    ----------
    encoding module
    """
    conv = tfp.layers.Convolution3DFlipout(
        filters,
        kernel_size,
        activation=activation,
        padding=padding,
        kernel_divergence_fn=kld,
        kernel_posterior_fn=kernel_posterior_fn,
        kernel_prior_fn=prior_fn,
    )(inputs)
    conv = GroupNormalization()(conv)
    conv = tfp.layers.Convolution3DFlipout(
        filters,
        kernel_size,
        activation=activation,
        padding=padding,
        kernel_divergence_fn=kld,
        kernel_posterior_fn=kernel_posterior_fn,
        kernel_prior_fn=prior_fn,
    )(conv)
    conv = GroupNormalization()(conv)
    pool = MaxPooling3D()(conv)
    return conv, pool


def up_stage(
    inputs,
    skip,
    filters,
    prior_fn,
    kernel_posterior_fn,
    kld,
    kernel_size=3,
    activation="relu",
    padding="SAME",
):
    """decoding blocks of the Bayesian VNet model.

    Parameters
    ----------
    inputs: tf.layer for encoding stage.
    skip: setting skip connections
    kld: a func to compute KL Divergence loss, default is set None.
        KLD can be set as (lambda q, p, ignore: kl_lib.kl_divergence(q, p))
    prior_fn: a func to initialize priors distributions
    kernel_posterior_fn:a func to initlaize kernel posteriors
        (loc, scale and weightnorms)
    filters: list or tuple of four ints, the shape of the input data. Omit
        the batch dimension, and include the number of channels.
    kernal_size: int, size of the kernel of conv layers. Default kernel size
        is set to be 3.
    activation: str or optimizer object, the non-linearity to use. All
        tf.activations are allowed to use

    Returns
    ----------
    decoding module
    """
    up = UpSampling3D()(inputs)
    up = tfp.layers.Convolution3DFlipout(
        filters,
        2,
        activation=activation,
        padding=padding,
        kernel_divergence_fn=kld,
        kernel_posterior_fn=kernel_posterior_fn,
        kernel_prior_fn=prior_fn,
    )(up)
    up = GroupNormalization()(up)

    merge = concatenate([skip, up])
    merge = GroupNormalization()(merge)

    conv = tfp.layers.Convolution3DFlipout(
        filters,
        kernel_size,
        activation=activation,
        padding=padding,
        kernel_divergence_fn=kld,
        kernel_posterior_fn=kernel_posterior_fn,
        kernel_prior_fn=prior_fn,
    )(merge)
    conv = GroupNormalization()(conv)
    conv = tfp.layers.Convolution3DFlipout(
        filters,
        kernel_size,
        activation=activation,
        padding=padding,
        kernel_divergence_fn=kld,
        kernel_posterior_fn=kernel_posterior_fn,
        kernel_prior_fn=prior_fn,
    )(conv)
    conv = GroupNormalization()(conv)

    return conv


def end_stage(
    inputs,
    prior_fn,
    kernel_posterior_fn,
    kld,
    n_classes=1,
    kernel_size=3,
    activation="relu",
    padding="SAME",
):
    """Last logit layer.

    Parameters
    ----------
    inputs: tf.model layer.
    kld: a func to compute KL Divergence loss, default is set None.
        KLD can be set as (lambda q, p, ignore: kl_lib.kl_divergence(q, p))
    prior_fn: a func to initialize priors distributions
    kernel_posterior_fn:a func to initlaize kernel posteriors
        (loc, scale and weightnorms)
    n_classes: int, for binary class use the value 1.
    kernal_size: int, size of the kernel of conv layers. Default kernel size
        is set to be 3.
    activation: str or optimizer object, the non-linearity to use. All
        tf.activations are allowed to use

    Result
    ----------
    prediction probabilities
    """
    conv = tfp.layers.Convolution3DFlipout(
        n_classes,
        kernel_size,
        activation=activation,
        padding="SAME",
        kernel_divergence_fn=kld,
        kernel_posterior_fn=kernel_posterior_fn,
        kernel_prior_fn=prior_fn,
    )(inputs)
    if n_classes == 1:
        conv = tfp.layers.Convolution3DFlipout(
            n_classes,
            1,
            activation="sigmoid",
            kernel_divergence_fn=kld,
            kernel_posterior_fn=kernel_posterior_fn,
            kernel_prior_fn=prior_fn,
        )(conv)
    else:
        conv = tfp.layers.Convolution3DFlipout(
            n_classes,
            1,
            activation="softmax",
            kernel_divergence_fn=kld,
            kernel_posterior_fn=kernel_posterior_fn,
            kernel_prior_fn=prior_fn,
        )(conv)
    return conv


def bayesian_vnet(
    n_classes=1,
    input_shape=(280, 280, 280, 1),
    kernel_size=3,
    prior_fn=normal_prior(),
    kernel_posterior_fn=tfp.layers.default_mean_field_normal_fn(),
    kld=None,
    activation="relu",
    padding="SAME",
):
    """Instantiate a 3D Bayesian VNet Architecture.

    Adapted from Deterministic VNet: https://arxiv.org/pdf/1606.04797.pdf
    Encoder and Decoder has 3D Flipout(variational layers)

    Parameters
    ----------
    n_classes: int, number of classes to classify. For binary applications, use
        a value of 1.
    input_shape: list or tuple of four ints, the shape of the input data. Omit
        the batch dimension, and include the number of channels.
    kernal_size(int): size of the kernel of conv layers
    activation(str): all tf.keras.activations are allowed
    kld: a func to compute KL Divergence loss, default is set None.
        KLD can be set as (lambda q, p, ignore: kl_lib.kl_divergence(q, p))
    prior_fn: a func to initialize priors distributions
    kernel_posterior_fn:a func to initlaize kernel posteriors
        (loc, scale and weightnorms)
    See Bayesian Utils for more options for kld, prior_fn and kernal_posterior_fn
    activation: str or optimizer object, the non-linearity to use. All
        tf.activations are allowed to use.

    Returns
    ----------
    Bayesian model object.
    """
    inputs = Input(input_shape)

    conv1, pool1 = down_stage(
        inputs,
        16,
        prior_fn,
        kernel_posterior_fn,
        kld,
        kernel_size=kernel_size,
        activation=activation,
        padding=padding,
    )
    conv2, pool2 = down_stage(
        pool1,
        32,
        prior_fn,
        kernel_posterior_fn,
        kld,
        kernel_size=kernel_size,
        activation=activation,
        padding=padding,
    )
    conv3, pool3 = down_stage(
        pool2,
        64,
        prior_fn,
        kernel_posterior_fn,
        kld,
        kernel_size=kernel_size,
        activation=activation,
        padding=padding,
    )
    conv4, _ = down_stage(
        pool3,
        128,
        prior_fn,
        kernel_posterior_fn,
        kld,
        kernel_size=kernel_size,
        activation=activation,
        padding=padding,
    )

    conv5 = up_stage(
        conv4,
        conv3,
        64,
        prior_fn,
        kernel_posterior_fn,
        kld,
        kernel_size=kernel_size,
        activation=activation,
        padding=padding,
    )
    conv6 = up_stage(
        conv5,
        conv2,
        32,
        prior_fn,
        kernel_posterior_fn,
        kld,
        kernel_size=kernel_size,
        activation=activation,
        padding=padding,
    )
    conv7 = up_stage(
        conv6,
        conv1,
        16,
        prior_fn,
        kernel_posterior_fn,
        kld,
        kernel_size=kernel_size,
        activation=activation,
        padding=padding,
    )

    conv8 = end_stage(
        conv7,
        prior_fn,
        kernel_posterior_fn,
        kld,
        n_classes=n_classes,
        kernel_size=kernel_size,
        activation=activation,
        padding=padding,
    )

    return Model(inputs=inputs, outputs=conv8)



import random

import numpy as np
from numpy.testing import assert_array_equal
import pytest
import tensorflow as tf

from .utils import csv_of_volumes  # noqa: F401
from .. import io, tfrecord


def test_write_read_volume_labels(csv_of_volumes, tmp_path):  # noqa: F811
    files = io.read_csv(csv_of_volumes, skip_header=False)
    filename_template = str(tmp_path / "data-{shard:03d}.tfrecords")
    examples_per_shard = 12
    tfrecord.write(
        files,
        filename_template=filename_template,
        examples_per_shard=examples_per_shard,
        processes=1,
    )

    paths = list(tmp_path.glob("data-*.tfrecords"))
    paths = sorted(paths)
    assert len(paths) == 9
    assert (tmp_path / "data-008.tfrecords").is_file()

    dset = tf.data.TFRecordDataset(list(map(str, paths)), compression_type="GZIP")
    dset = dset.map(
        tfrecord.parse_example_fn(volume_shape=(8, 8, 8), scalar_labels=False)
    )

    for ref, test in zip(files, dset):
        x, y = ref
        x, y = io.read_volume(x), io.read_volume(y)
        assert_array_equal(x, test[0])
        assert_array_equal(y, test[1])

    with pytest.raises(ValueError):
        tfrecord.write(
            files, filename_template="data/foobar-{}.tfrecords", examples_per_shard=4
        )


def test_write_read_volume_labels_all_processes(csv_of_volumes, tmp_path):  # noqa: F811
    files = io.read_csv(csv_of_volumes, skip_header=False)
    filename_template = str(tmp_path / "data-{shard:03d}.tfrecords")
    examples_per_shard = 12
    tfrecord.write(
        files,
        filename_template=filename_template,
        examples_per_shard=examples_per_shard,
        processes=None,
    )

    paths = list(tmp_path.glob("data-*.tfrecords"))
    paths = sorted(paths)
    assert len(paths) == 9
    assert (tmp_path / "data-008.tfrecords").is_file()

    dset = tf.data.TFRecordDataset(list(map(str, paths)), compression_type="GZIP")
    dset = dset.map(
        tfrecord.parse_example_fn(volume_shape=(8, 8, 8), scalar_labels=False)
    )

    for ref, test in zip(files, dset):
        x, y = ref
        x, y = io.read_volume(x), io.read_volume(y)
        assert_array_equal(x, test[0])
        assert_array_equal(y, test[1])

    with pytest.raises(ValueError):
        tfrecord.write(
            files, filename_template="data/foobar-{}.tfrecords", examples_per_shard=4
        )


def test_write_read_float_labels(csv_of_volumes, tmp_path):  # noqa: F811
    files = io.read_csv(csv_of_volumes, skip_header=False)
    files = [(x, random.random()) for x, _ in files]
    filename_template = str(tmp_path / "data-{shard:03d}.tfrecords")
    examples_per_shard = 12
    tfrecord.write(
        files,
        filename_template=filename_template,
        examples_per_shard=examples_per_shard,
        processes=1,
    )

    paths = list(tmp_path.glob("data-*.tfrecords"))
    paths = sorted(paths)
    assert len(paths) == 9
    assert (tmp_path / "data-008.tfrecords").is_file()

    dset = tf.data.TFRecordDataset(list(map(str, paths)), compression_type="GZIP")
    dset = dset.map(
        tfrecord.parse_example_fn(volume_shape=(8, 8, 8), scalar_labels=True)
    )

    for ref, test in zip(files, dset):
        x, y = ref
        x = io.read_volume(x)
        assert_array_equal(x, test[0])
        assert_array_equal(y, test[1])


def test_write_read_int_labels(csv_of_volumes, tmp_path):  # noqa: F811
    files = io.read_csv(csv_of_volumes, skip_header=False)
    files = [(x, random.randint(0, 9)) for x, _ in files]
    filename_template = str(tmp_path / "data-{shard:03d}.tfrecords")
    examples_per_shard = 12
    tfrecord.write(
        files,
        filename_template=filename_template,
        examples_per_shard=examples_per_shard,
        processes=1,
    )

    paths = list(tmp_path.glob("data-*.tfrecords"))
    paths = sorted(paths)
    assert len(paths) == 9
    assert (tmp_path / "data-008.tfrecords").is_file()

    dset = tf.data.TFRecordDataset(list(map(str, paths)), compression_type="GZIP")
    dset = dset.map(
        tfrecord.parse_example_fn(volume_shape=(8, 8, 8), scalar_labels=True)
    )

    for ref, test in zip(files, dset):
        x, y = ref
        x = io.read_volume(x)
        assert_array_equal(x, test[0])
        assert_array_equal(y, test[1])


def test__is_int_or_float():
    assert tfrecord._is_int_or_float(10)
    assert tfrecord._is_int_or_float(10.0)
    assert tfrecord._is_int_or_float("10")
    assert tfrecord._is_int_or_float("10.00")
    assert not tfrecord._is_int_or_float("foobar")
    assert tfrecord._is_int_or_float(np.ones(1))
    assert not tfrecord._is_int_or_float(np.ones(10))


def test__dtype_to_bytes():
    np_tf_dt = [
        (np.uint8, tf.uint8, b"uint8"),
        (np.uint16, tf.uint16, b"uint16"),
        (np.uint32, tf.uint32, b"uint32"),
        (np.uint64, tf.uint64, b"uint64"),
        (np.int8, tf.int8, b"int8"),
        (np.int16, tf.int16, b"int16"),
        (np.int32, tf.int32, b"int32"),
        (np.int64, tf.int64, b"int64"),
        (np.float16, tf.float16, b"float16"),
        (np.float32, tf.float32, b"float32"),
        (np.float64, tf.float64, b"float64"),
    ]

    for npd, tfd, dt in np_tf_dt:
        npd = np.dtype(npd)
        assert tfrecord._dtype_to_bytes(npd) == dt
        assert tfrecord._dtype_to_bytes(tfd) == dt

    assert tfrecord._dtype_to_bytes("float32") == b"float32"
    assert tfrecord._dtype_to_bytes("foobar") == b"foobar"



import importlib
import logging

import tensorflow as tf

from .base import BaseEstimator
from .. import losses, metrics
from ..models import available_models, list_available_models

logging.getLogger().setLevel(logging.INFO)


class Segmentation(BaseEstimator):
    """Perform segmentation type operations"""

    state_variables = ["block_shape_", "volume_shape_", "scalar_labels_"]

    def __init__(
        self, base_model, model_args=None, checkpoint_filepath=None, multi_gpu=True
    ):
        super().__init__(checkpoint_filepath=checkpoint_filepath, multi_gpu=multi_gpu)

        if not isinstance(base_model, str):
            self.base_model = base_model.__name__
        else:
            self.base_model = base_model

        if self.base_model and self.base_model not in available_models():
            raise ValueError(
                "Unknown model: '{}'. Available models are {}.".format(
                    self.base_model, available_models()
                )
            )

        self.model_ = None
        self.model_args = model_args or {}
        self.block_shape_ = None
        self.volume_shape_ = None
        self.scalar_labels_ = None

    def add_model(self, base_model, model_args=None):
        """Add a segmentation model"""
        self.base_model = base_model
        self.model_args = model_args or {}

    def fit(
        self,
        dataset_train,
        dataset_validate=None,
        epochs=1,
        # TODO: figure out whether optimizer args should be flattened
        optimizer=None,
        opt_args=None,
        loss=losses.dice,
        metrics=metrics.dice,
        callbacks=None,
        verbose=1,
    ):
        """Train a segmentation model"""
        # TODO: check validity of datasets

        batch_size = dataset_train.batch_size
        self.block_shape_ = dataset_train.block_shape
        self.volume_shape_ = dataset_train.volume_shape
        self.scalar_labels_ = dataset_train.scalar_labels
        n_classes = dataset_train.n_classes
        opt_args = opt_args or {}
        if optimizer is None:
            optimizer = tf.keras.optimizers.Adam
            opt_args_tmp = dict(learning_rate=1e-04)
            opt_args_tmp.update(**opt_args)
            opt_args = opt_args_tmp

        def _create(base_model):
            # Instantiate and compile the model
            self.model_ = base_model(
                n_classes=n_classes,
                input_shape=(*self.block_shape_, 1),
                **self.model_args
            )

        def _compile():
            self.model_.compile(
                optimizer(**opt_args),
                loss=loss,
                metrics=metrics,
            )

        if self.model is None:
            mod = importlib.import_module("..models", "nobrainer.processing")
            base_model = getattr(mod, self.base_model)
            if batch_size % self.strategy.num_replicas_in_sync:
                raise ValueError("batch size must be a multiple of the number of GPUs")

            with self.strategy.scope():
                _create(base_model)
        with self.strategy.scope():
            _compile()
        self.model_.summary()

        if callbacks is not None and not isinstance(callbacks, list):
            raise AttributeError("Callbacks must be either of type list or None")

        if callbacks is None:
            callbacks = []

        if self.checkpoint_tracker:
            callbacks.append(self.checkpoint_tracker)
        self.model_.fit(
            dataset_train.dataset,
            epochs=epochs,
            steps_per_epoch=dataset_train.get_steps_per_epoch(),
            validation_data=dataset_validate.dataset if dataset_validate else None,
            validation_steps=(
                dataset_validate.get_steps_per_epoch() if dataset_validate else None
            ),
            callbacks=callbacks,
            verbose=verbose,
        )

        return self

    def predict(self, x, batch_size=1, normalizer=None):
        """Makes a prediction using the trained model"""
        if self.model_ is None:
            raise ValueError("Model is undefined. Please train or load a model")
        from ..prediction import predict

        return predict(
            x,
            self.model_,
            block_shape=self.block_shape_,
            batch_size=batch_size,
            normalizer=normalizer,
        )

    @classmethod
    def list_available_models(cls):
        list_available_models()



import os
from pathlib import Path

import tensorflow as tf

from .base import BaseEstimator
from .. import losses
from ..dataset import get_dataset


class ProgressiveGeneration(BaseEstimator):
    """Perform generation type operations"""

    state_variables = ["current_resolution_"]

    def __init__(
        self,
        latent_size=256,
        label_size=0,
        num_channels=1,
        dimensionality=3,
        g_fmap_base=1024,
        d_fmap_base=1024,
        multi_gpu=True,
    ):
        super().__init__(multi_gpu=multi_gpu)
        self.model_ = None
        self.latent_size = latent_size
        self.label_size = label_size
        self.g_fmap_base = g_fmap_base
        self.d_fmap_base = d_fmap_base
        self.num_channels = num_channels
        self.dimensionality = dimensionality
        self.current_resolution_ = 0

    def fit(
        self,
        dataset_train,
        epochs=2,
        checkpoint_dir=Path(os.getcwd()) / "temp",
        normalizer=None,
        # TODO: figure out whether optimizer args should be flattened
        g_optimizer=None,
        g_opt_args=None,
        d_optimizer=None,
        d_opt_args=None,
        g_loss=losses.Wasserstein,
        d_loss=losses.Wasserstein,
        warm_start=False,
        num_parallel_calls=None,
        save_freq=500,
    ):
        """Train a progressive gan model"""
        # TODO: check validity of datasets

        # create checkpoint sub-dirs
        checkpoint_dir = Path(checkpoint_dir)
        checkpoint_dir.mkdir(exist_ok=True)
        generated_dir = checkpoint_dir / "generated"
        model_dir = checkpoint_dir / "saved_models"
        log_dir = checkpoint_dir / "logs"

        generated_dir.mkdir(exist_ok=True)
        model_dir.mkdir(exist_ok=True)
        log_dir.mkdir(exist_ok=True)

        # set optimizers
        g_opt_args = g_opt_args or {}
        if g_optimizer is None:
            g_optimizer = tf.keras.optimizers.Adam
        g_opt_args_tmp = dict(
            learning_rate=1e-04, beta_1=0.0, beta_2=0.99, epsilon=1e-8
        )
        g_opt_args_tmp.update(**g_opt_args)
        g_opt_args = g_opt_args_tmp

        d_opt_args = d_opt_args or {}
        if d_optimizer is None:
            d_optimizer = tf.keras.optimizers.Adam
        d_opt_args_tmp = dict(
            learning_rate=1e-04, beta_1=0.0, beta_2=0.99, epsilon=1e-8
        )
        d_opt_args_tmp.update(**d_opt_args)
        d_opt_args = d_opt_args_tmp

        if warm_start:
            if self.model_ is None:
                raise ValueError("warm_start requested, but model is undefined")
        else:
            from ..models.progressivegan import progressivegan
            from ..training import ProgressiveGANTrainer

            # Instantiate the generator and discriminator
            with self.strategy.scope():
                generator, discriminator = progressivegan(
                    latent_size=self.latent_size,
                    g_fmap_base=self.g_fmap_base,
                    d_fmap_base=self.d_fmap_base,
                    num_channels=self.num_channels,
                    dimensionality=self.dimensionality,
                )
                self.model_ = ProgressiveGANTrainer(
                    generator=generator,
                    discriminator=discriminator,
                    gradient_penalty=True,
                )
            self.current_resolution_ = 0

        # wrap the losses to work on multiple GPUs
        with self.strategy.scope():
            d_loss_object = d_loss(reduction=tf.keras.losses.Reduction.NONE)

            def compute_d_loss(labels, predictions):
                per_example_loss = d_loss_object(labels, predictions)
                return tf.nn.compute_average_loss(
                    per_example_loss, global_batch_size=batch_size
                )

            g_loss_object = g_loss(reduction=tf.keras.losses.Reduction.NONE)

            def compute_g_loss(labels, predictions):
                per_example_loss = g_loss_object(labels, predictions)
                return tf.nn.compute_average_loss(
                    per_example_loss, global_batch_size=batch_size
                )

            d_loss = compute_d_loss
            g_loss = compute_g_loss

        # instantiate a progressive training helper and compile with loss and optimizer
        def _compile():
            self.model_.compile(
                g_optimizer=g_optimizer(**g_opt_args),
                d_optimizer=d_optimizer(**d_opt_args),
                g_loss_fn=g_loss,
                d_loss_fn=d_loss,
            )

        print(self.model_.generator.summary())
        print(self.model_.discriminator.summary())

        for resolution, info in dataset_train.items():
            if resolution < self.current_resolution_:
                continue
            # create a train dataset with features for resolution
            batch_size = info.get("batch_size")
            if batch_size % self.strategy.num_replicas_in_sync:
                raise ValueError("batch size must be a multiple of the number of GPUs")

            dataset = get_dataset(
                file_pattern=info.get("file_pattern"),
                batch_size=batch_size,
                num_parallel_calls=num_parallel_calls,
                volume_shape=(resolution, resolution, resolution),
                n_classes=1,
                scalar_label=True,
                normalizer=info.get("normalizer") or normalizer,
            )

            with self.strategy.scope():
                # grow the networks by one (2^x) resolution
                if resolution > self.current_resolution_:
                    self.model_.generator.add_resolution()
                    self.model_.discriminator.add_resolution()
                _compile()

                steps_per_epoch = (info.get("epochs") or epochs) // info.get(
                    "batch_size"
                )

                # save_best_only is set to False as it is an adversarial loss
                model_checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
                    str(model_dir),
                    save_weights_only=True,
                    save_best_only=False,
                    save_freq=save_freq,
                    verbose=False,
                )

            # Train at resolution
            print("Resolution : {}".format(resolution))

            print("Transition phase")
            self.model_.fit(
                dataset,
                phase="transition",
                resolution=resolution,
                steps_per_epoch=steps_per_epoch,  # necessary for repeat dataset
                callbacks=[model_checkpoint_callback],
            )

            print("Resolution phase")
            self.model_.fit(
                dataset,
                phase="resolution",
                resolution=resolution,
                steps_per_epoch=steps_per_epoch,
                callbacks=[model_checkpoint_callback],
            )
            self.current_resolution_ = resolution
            # save the final weights
            self.model_.save_weights(model_dir)
        return self

    def generate(self, n_images=1, return_latents=False, data_type=None):
        """generate a synthetic image using the trained model"""
        if self.model_ is None:
            raise ValueError("Model is undefined. Please train or load a model")
        import nibabel as nib
        import numpy as np

        latents_all = []
        img_all = []
        for i in range(n_images):
            latents = tf.random.normal((1, self.latent_size))
            img = self.model_.generator.generate(latents)["generated"]
            img = np.squeeze(img)
            if data_type is not None:
                img = np.round(
                    np.iinfo(data_type).max
                    * (img - img.min())
                    / (img.max() - img.min())
                ).astype(data_type)
            img = nib.Nifti1Image(img, np.eye(4))
            latents_all.append(latents)
            img_all.append(img)
        if return_latents:
            return img_all, latents_all
        else:
            return img_all
